searchData={"items":[{"type":"behaviour","title":"Slipstream","doc":"A websocket client for Phoenix channels\n\nSlipstream is a websocket client for connection to `Phoenix.Channel`s.\nSlipstream is a bit different from existing websocket implementations in that:\n\n- it's backed by `Mint.WebSocket`\n- it has an `await_*` interface for performing actions synchronously\n- smart retry strategies for reconnection and rejoining work out-of-the-box\n- a testing framework for clients\n- high-level and low-level instrumentation with `:telemetry`","ref":"Slipstream.html"},{"type":"behaviour","title":"Basic Usage - Slipstream","doc":"The intended use for Slipstream is to write asynchronous, callback-oriented\nGenServer-like modules that define socket clients. This approach makes it\neasy to write socket clients that resemble state-machines. A minimalistic\nexample usage might be like so:\n\n    defmodule MyApp.MySocketClient do\n      @moduledoc \"\"\"\n      A socket client for connecting to that other Phoenix server\n\n      Periodically sends pings and asks the other server for its metrics.\n      \"\"\"\n\n      use Slipstream,\n        restart: :temporary\n\n      require Logger\n\n      @topic \"backend-service:money-server\"\n\n      def start_link(args) do\n        Slipstream.start_link(__MODULE__, args, name: __MODULE__)\n      end\n\n      @impl Slipstream\n      def init(config) do\n        {:ok, connect!(config), {:continue, :start_ping}}\n      end\n\n      @impl Slipstream\n      def handle_continue(:start_ping, socket) do\n        timer = :timer.send_interval(1000, self(), :request_metrics)\n\n        {:noreply, assign(socket, :ping_timer, timer)}\n      end\n\n      @impl Slipstream\n      def handle_connect(socket) do\n        {:ok, join(socket, @topic)}\n      end\n\n      @impl Slipstream\n      def handle_join(@topic, _join_response, socket) do\n        # an asynchronous push with no reply:\n        push(socket, @topic, \"hello\", %{})\n\n        {:ok, socket}\n      end\n\n      @impl Slipstream\n      def handle_info(:request_metrics, socket) do\n        # we will asynchronously receive a reply and handle it in the\n        # handle_reply/3 implementation below\n        {:ok, ref} = push(socket, @topic, \"get_metrics\", %{format: \"json\"})\n\n        {:noreply, assign(socket, :metrics_request, ref)}\n      end\n\n      @impl Slipstream\n      def handle_reply(ref, metrics, socket) do\n        if ref == socket.assigns.metrics_request do\n          :ok = MyApp.MetricsPublisher.publish(metrics)\n        end\n\n        {:ok, socket}\n      end\n\n      @impl Slipstream\n      def handle_message(@topic, event, message, socket) do\n        Logger.error(\n          \"Was not expecting a push from the server. Heard: \" <>\n            inspect({@topic, event, message})\n        )\n\n        {:ok, socket}\n      end\n\n      @impl Slipstream\n      def handle_disconnect(_reason, socket) do\n        :timer.cancel(socket.assigns.ping_timer)\n\n        {:stop, :normal, socket}\n      end\n    end","ref":"Slipstream.html#module-basic-usage"},{"type":"behaviour","title":"Synchronicity - Slipstream","doc":"Slipstream is designed to work asynchronously by default. Requests such\nas `connect/2`, `join/3`, and `push/4` are asynchronous requests. When\nthe remote server replies, the associated callback will be invoked\n(`c:handle_connect/1`, `c:handle_join/3`, and `c:handle_reply/3` in cases\nof success, respectively). For all of these operations, though, you may\nawait the outcome of the asynchronous request with `await_*` functions. E.g.\n\n    iex> {:ok, ref} = push(socket, \"room:lobby\", \"msg:new\", %{user: 1, msg: \"foo\"})\n    iex> {:ok, %{\"created\" => true}} = await_reply(ref)\n\nNote that all `await_*` functions must be called from the slipstream process\nthat emitted the request, or else they will timeout.\n\nWhile Slipstream provides a rich toolset for synchronicity, the asynchronous,\ncallback-based workflow is recommended.","ref":"Slipstream.html#module-synchronicity"},{"type":"behaviour","title":"GenServer operations - Slipstream","doc":"Note that Slipstream is in many ways a simple wrapper around a GenServer.\nAs such, all GenServer functionality is possible with Slipstream clients,\nsuch as `Kernel.send/2` or `GenServer.call/3`. For example, assume you have\na slipstream client written like so:\n\n    defmodule MyClient do\n      use Slipstream\n\n      require Logger\n\n      def start_link(args) do\n        Slipstream.start_link(__MODULE__, args, name: __MODULE__)\n      end\n\n      @impl Slipstream\n      def init(config), do: connect(config)\n\n      @impl Slipstream\n      def handle_cast(:ping, socket) do\n        Logger.info(\"pong\")\n\n        {:noreply, socket}\n      end\n\n      @impl Slipstream\n      def handle_info(:hello, socket) do\n        Logger.info(\"hello\")\n\n        {:noreply, socket}\n      end\n\n      @impl Slipstream\n      def handle_call(:foo, _from, socket) do\n        {:reply, {:ok, :bar}, socket}\n      end\n\n      ..\n    end\n\nThis `MyClient` client is a GenServer, so the following are valid ways to\ninteract with `MyClient`:\n\n    iex> GenServer.cast(MyClient, :ping)\n    [info] pong\n    :ok\n    iex> MyClient |> GenServer.whereis |> send(:hello)\n    [info] hello\n    :hello\n    iex> GenServer.call(MyClient, :foo)\n    {:ok, :bar}","ref":"Slipstream.html#module-genserver-operations"},{"type":"behaviour","title":"Retry Mechanisms - Slipstream","doc":"There are two levels at which retry logic may be needed: for maintaining a\nwebsocket connection and for retrying messages to which the server replies\nwith an error or does not reply.\nSlipstream does not implement message retries because the needs of applications\nwill vary; if you need retries, you can build the logic you need using\n`push/4`, `c:handle_reply/3`, and something like `Process.send_after/4`.\n\nTo maintain a websocket connection, Slipstream emulates the official\n`phoenix.js` package with its reconnection and re-join features.\n`Slipstream.Configuration` allows configuration of the back-off times with\nthe `:reconnect_after_msec` and `:rejoin_after_msec` lists, respectively.\n\nTo take advantage of these built-in mechanisms, a client must be written\nin the asynchronous GenServer-like manner and must use the `reconnect/1` and\n`rejoin/3` functions in its `c:Slipstream.handle_disconnect/2` and\n`c:Slipstream.handle_topic_close/3` callbacks, respectively. Note that the\ndefault implementation of these callbacks invokes these functions, so a client\nwhich does not explicitly define these callbacks will retry connection and\njoins.\n\nTake care to handle the `:left` case of `c:Slipstream.handle_topic_close/3`.\nIn the case that a client attempts to leave a topic with `leave/2`, the\ncallback will be invoked with a `reason` of `:left`. The default\nimplementation of `c:Slipstream.handle_topic_close/3` makes this distinction\nand simply no-ops on channel leaves.\n\n    defmodule MyClientWithRetry do\n      use Slipstream\n\n      def start_link(config) do\n        Slipstream.start_link(__MODULE__, config, name: __MODULE__)\n      end\n\n      @impl Slipstream\n      def init(config), do: connect(config)\n\n      @impl Slipstream\n      def handle_connect(socket) do\n        {:ok, join(socket, \"rooms:lobby\", %{user_id: 1})}\n      end\n\n      @impl Slipstream\n      def handle_disconnect(_reason, socket) do\n        case reconnect(socket) do\n          {:ok, socket} -> {:ok, socket}\n          {:error, reason} -> {:stop, reason, socket}\n        end\n      end\n\n      @impl Slipstream\n      def handle_topic_close(topic, _reason, socket) do\n        rejoin(socket, topic)\n      end\n    end","ref":"Slipstream.html#module-retry-mechanisms"},{"type":"macro","title":"Slipstream.__using__/1","doc":"Declares that a module is a Slipstream socket client\n\nSlipstream provides a `use Slipstream` macro that behaves similar to\nGenServer's `use GenServer`. This does a few things:\n\n- a default implementation of `child_spec/1`, which is used to start the\n  module as a GenServer. This may be overridden.\n- imports for all documented functions in `Slipstream` and `Slipstream.Socket`\n- a `c:GenServer.handle_info/2` function clause which matches incoming events\n  from the connection process and dispatches them to the various Slipstream\n  callbacks\n- a `c:GenServer.handle_info/2` function clause which matches Slipstream\n  commands. This is used to implement back-off retry mechanisms for\n  `reconnect/1` and `rejoin/3`.\n\nThis provides a familiar and sleek interface for the common case of using\nSlipstream: an asynchronous callback-based GenServer module.\n\nIt's not required to use this macro, though. Slipstream can be used in\nsynchronous mode (via the `await_*` family of functions).","ref":"Slipstream.html#__using__/1"},{"type":"macro","title":"Options - Slipstream.__using__/1","doc":"Any options passed as the `opts` argument to `__using__/1` are passed along\nto `Supervisor.child_spec/2`, as is done in the default `child_spec/1`\nimplementation for GenServer. Most notably, you may control the restart\nstrategy of the client with the `:restart` option. See the example below.","ref":"Slipstream.html#__using__/1-options"},{"type":"macro","title":"Examples - Slipstream.__using__/1","doc":"defmodule MyApp.MySocketClient do\n      # one crash/shutdown/exit will permanently terminate the server\n      use Slipstream, restart: :temporary\n\n      def start_link(args) do\n        Slipstream.start_link(__MODULE__, args)\n      end\n\n      ..\n    end","ref":"Slipstream.html#__using__/1-examples"},{"type":"function","title":"Slipstream.await_connect/2","doc":"Awaits a pending connection request synchronously","ref":"Slipstream.html#await_connect/2"},{"type":"function","title":"Slipstream.await_connect!/2","doc":"Awaits a pending connection request synchronously, raising on failure","ref":"Slipstream.html#await_connect!/2"},{"type":"function","title":"Slipstream.await_disconnect/2","doc":"Awaits a pending disconnection request synchronously","ref":"Slipstream.html#await_disconnect/2"},{"type":"function","title":"Slipstream.await_disconnect!/2","doc":"Awaits a pending disconnection request synchronously, raising on failure","ref":"Slipstream.html#await_disconnect!/2"},{"type":"function","title":"Slipstream.await_join/3","doc":"Awaits a pending join request synchronously","ref":"Slipstream.html#await_join/3"},{"type":"function","title":"Slipstream.await_join!/3","doc":"Awaits a join request synchronously, raising on failure","ref":"Slipstream.html#await_join!/3"},{"type":"function","title":"Slipstream.await_leave/3","doc":"Awaits a leave request synchronously","ref":"Slipstream.html#await_leave/3"},{"type":"function","title":"Slipstream.await_leave!/3","doc":"Awaits a leave request synchronously, raising on failure","ref":"Slipstream.html#await_leave!/3"},{"type":"macro","title":"Slipstream.await_message/4","doc":"Awaits the server's message push\n\nNote that unlike the other `await_*` functions, `await_message/4` is a\nmacro. This allows an author to match on patterns in the topic, event, and/or\npayload parts of a message.","ref":"Slipstream.html#await_message/4"},{"type":"macro","title":"Examples - Slipstream.await_message/4","doc":"iex> event = \"msg:new\"\n    iex> await_message(\"rooms:lobby\", ^event, %{\"user_id\" => 5})\n    {:ok, \"rooms:lobby\", \"msg:new\", %{\"user_id\" => 5, body: \"hello\"}}","ref":"Slipstream.html#await_message/4-examples"},{"type":"macro","title":"Slipstream.await_message!/4","doc":"Awaits the server's message push, raising on failure\n\nSee `await_message/4`","ref":"Slipstream.html#await_message!/4"},{"type":"function","title":"Slipstream.await_reply/2","doc":"Awaits the server's response to a message","ref":"Slipstream.html#await_reply/2"},{"type":"function","title":"Slipstream.await_reply!/2","doc":"Awaits the server's response to a message, exiting on timeout\n\nSee `await_reply/2` for more information.","ref":"Slipstream.html#await_reply!/2"},{"type":"function","title":"Slipstream.collect_garbage/1","doc":"Requests that the connection process undergoe garbage collection\n\nIf you're using Slipstream to send large messages, you may wish to flush\nthe process memory of the connection process between large messages. This\ncan be achieved through garbage collection.","ref":"Slipstream.html#collect_garbage/1"},{"type":"function","title":"Examples - Slipstream.collect_garbage/1","doc":"iex> collect_garbage(socket)\n    :ok","ref":"Slipstream.html#collect_garbage/1-examples"},{"type":"function","title":"Slipstream.connect/2","doc":"Requests connection to the remote endpoint\n\n`opts` are passed to `Slipstream.Configuration.validate/1` before sending.\n\nNote that this request for connection is asynchronous. A return value of\n`{:ok, socket}` does not mean that a connection has successfully been\nestablished.","ref":"Slipstream.html#connect/2"},{"type":"function","title":"Examples - Slipstream.connect/2","doc":"{:ok, socket} = connect(uri: \"ws://localhost:4000/socket/websocket\")","ref":"Slipstream.html#connect/2-examples"},{"type":"function","title":"Slipstream.connect!/2","doc":"Same as `connect/2` but raises on configuration validation error\n\nNote that `connect!/2` will not necessarily raise an error on failure to\nconnect. The `!` only pertains to the potential for raising when the\nconfiguration is invalid.","ref":"Slipstream.html#connect!/2"},{"type":"function","title":"Slipstream.disconnect/1","doc":"Requests that an open connection be closed\n\nThis function will no-op when the socket is not currently connected to\nany remote websocket server.\n\nNote that you do not need to use `disconnect/1` to clean up a connection.\nThe connection process monitors the slipstream client process and will shut\ndown when it detects that the process has terminated.\n\nDisconnection may be awaited synchronously with `await_disconnect/2`","ref":"Slipstream.html#disconnect/1"},{"type":"function","title":"Examples - Slipstream.disconnect/1","doc":"@impl Slipstream\n    def handle_info(:chaos_monkey, socket) do\n      {:ok, socket} =\n        socket\n        |> disconnect()\n        |> await_disconnect()\n\n      {:noreply, reconnect(socket)}\n    end","ref":"Slipstream.html#disconnect/1-examples"},{"type":"callback","title":"Slipstream.handle_call/3","doc":"Invoked when a slipstream process receives a GenServer call\n\nBehaves the same as `c:GenServer.handle_call/3`","ref":"Slipstream.html#c:handle_call/3"},{"type":"callback","title":"Slipstream.handle_cast/2","doc":"Invoked when a slipstream process receives a GenServer cast\n\nBehaves the same as `c:GenServer.handle_cast/2`","ref":"Slipstream.html#c:handle_cast/2"},{"type":"callback","title":"Slipstream.handle_connect/1","doc":"Invoked when a connection has been established to a websocket server\n\nThis callback provides a good place to `join/3`.","ref":"Slipstream.html#c:handle_connect/1"},{"type":"callback","title":"Examples - Slipstream.handle_connect/1","doc":"@impl Slipstream\n    def handle_connect(socket) do\n      {:ok, socket}\n    end","ref":"Slipstream.html#c:handle_connect/1-examples"},{"type":"callback","title":"Slipstream.handle_continue/2","doc":"Invoked as a continuation of another GenServer callback\n\nGenServer callbacks may end with signatures that declare that the next\nfunction invoked should be a continuation. E.g.\n\n    def init(state) do\n      {:ok, state, {:continue, :my_continue}}\n    end\n\n    # this will be invoked immediately after `init/1`\n    def handle_continue(:my_continue, state) do\n      # do something with state\n\n      {:noreply, state}\n    end\n\nThis provides a way to schedule work to occur immediately after\nsuccessful initialization or to break work across multiple callbacks, which\ncan be useful for clients which are state-machine-like.\n\nSee `c:GenServer.handle_continue/2` for more information.","ref":"Slipstream.html#c:handle_continue/2"},{"type":"callback","title":"Slipstream.handle_disconnect/2","doc":"Invoked when a connection has been terminated\n\nThe default implementation of this callback requests reconnection","ref":"Slipstream.html#c:handle_disconnect/2"},{"type":"callback","title":"Examples - Slipstream.handle_disconnect/2","doc":"@impl Slipstream\n    def handle_disconnect(_reason, socket) do\n      case reconnect(socket) do\n        {:ok, socket} -> {:ok, socket}\n        {:error, reason} -> {:stop, reason, socket}\n      end\n    end","ref":"Slipstream.html#c:handle_disconnect/2-examples"},{"type":"callback","title":"Slipstream.handle_info/2","doc":"Invoked when a slipstream process receives a message\n\nBehaves the same as `c:GenServer.handle_info/2`","ref":"Slipstream.html#c:handle_info/2"},{"type":"callback","title":"Slipstream.handle_join/3","doc":"Invoked when the websocket server replies to the request to join","ref":"Slipstream.html#c:handle_join/3"},{"type":"callback","title":"Examples - Slipstream.handle_join/3","doc":"@impl Slipstream\n    def handle_join(\"rooms:echo\", %{}, socket) do\n      push(socket, \"rooms:echo\", \"echo\", %{\"ping\" => 1})\n\n      {:ok, socket}\n    end","ref":"Slipstream.html#c:handle_join/3-examples"},{"type":"callback","title":"Slipstream.handle_leave/2","doc":"Invoked when a join has concluded by a leave request\n\nThis callback is invoked when the remote server acknowledges that the client\nhas disconnected as a result of calling `leave/2`.\n\nThe default implementation of this callback performs a no-op.","ref":"Slipstream.html#c:handle_leave/2"},{"type":"callback","title":"Examples - Slipstream.handle_leave/2","doc":"@impl Slipstream\n    def handle_leave(topic, socket) do\n      Logger.info(\"Successfully left topic \" <> topic)\n\n      {:ok, socket}\n    end","ref":"Slipstream.html#c:handle_leave/2-examples"},{"type":"callback","title":"Slipstream.handle_message/4","doc":"Invoked when a message is received on the websocket connection\n\nThis callback will not be invoked for a message which is a reply. Those\nmessages will be handled in `c:handle_reply/3`.\n\nNote that while replying is supported on the server-side of the Phoenix\nChannel protocol, it is not supported by a client. Messages sent from\nthe server cannot be directly replied to.","ref":"Slipstream.html#c:handle_message/4"},{"type":"callback","title":"Examples - Slipstream.handle_message/4","doc":"@impl Slipstream\n    def handle_message(\"room:lobby\", \"new:msg\", params, socket) do\n      MyApp.Msg.create(params)\n\n      {:ok, socket}\n    end","ref":"Slipstream.html#c:handle_message/4-examples"},{"type":"callback","title":"Slipstream.handle_reply/3","doc":"Invoked when a message is received on the websocket connection which\nreferences a push from this client process.\n\n`ref` is the string reference returned from the `push/2` which resulted in\nthis reply.","ref":"Slipstream.html#c:handle_reply/3"},{"type":"callback","title":"Examples - Slipstream.handle_reply/3","doc":"@impl Slipstream\n    def handle_join(topic, _params, socket) do\n      my_req = push(socket, topic, \"msg:new\", %{\"foo\" => \"bar\"})\n\n      {:ok, assign(socket, :request, my_req)}\n    end\n\n    @impl Slipstream\n    def handle_reply(ref, reply, %{assigns: %{request: ref}} = socket) do\n      IO.inspect(reply, label: \"reply to my request\")\n\n      {:ok, socket}\n    end","ref":"Slipstream.html#c:handle_reply/3-examples"},{"type":"callback","title":"Slipstream.handle_topic_close/3","doc":"Invoked when a join has concluded\n\nThis callback will be invoked in the case that the remote channel crashes.\n`reason` is an error tuple `{:error, params :: json_serializable()}` where\n`params` is the message sent from the remote channel on crash.\n\nThe default implementation of this callback attempts to re-join the\nlast-joined topic.","ref":"Slipstream.html#c:handle_topic_close/3"},{"type":"callback","title":"Examples - Slipstream.handle_topic_close/3","doc":"@impl Slipstream\n    def handle_topic_close(topic, _reason, socket) do\n      {:ok, socket} = rejoin(socket, topic)\n    end","ref":"Slipstream.html#c:handle_topic_close/3-examples"},{"type":"callback","title":"Slipstream.init/1","doc":"Invoked when the slipstream process in started\n\nBehaves the same as `c:GenServer.init/1`, but the return state must be a\nnew `t:Slipstream.Socket.t/0`. Values from `c:init/1` that you'd\nlike to keep in state can be stored with `Slipstream.Socket.assign/3`.\n\nThis callback is a good place to request connection with `connect/2`. Note\nthat `connect/2` is an asynchronous request for connection. Awaiting\nconnection with `await_connect/2` is unwise in many scenarios, however,\nbecause failure to connect may result in an exit from the process, crashing\nthe supervision tree that started the process. If you wish to connect\nsynchronously upon init, a better approach could be:\n\n    @impl Slipstream\n    def init(_args) do\n      config = Application.fetch_env!(:my_app, __MODULE__)\n      socket = new_socket() |> assign(:connect_config, config)\n\n      {:ok, socket, {:continue, :connect}}\n    end\n\n    @impl Slipstream\n    def handle_continue(:connect, socket) do\n      {:ok, socket} = connect(socket, socket.assigns.connect_config)\n\n      {:ok, socket} = await_connect(socket)\n\n      {:noreply, socket}\n    end\n\nBut a more minimalistic approach that still provides safety in cases of\nconfiguration validation failures would be:\n\n    defmodule MySocketClient do\n      use Slipstream\n\n      def start_link(args) do\n        Slipstream.start_link(__MODULE__, args, name: __MODULE__)\n      end\n\n      @impl Slipstream\n      def init(_args) do\n        config = Application.fetch_env!(:my_app, __MODULE__)\n\n        case connect(config) do\n          {:ok socket} ->\n            {:ok, socket}\n\n          {:error, reason} ->\n            Logger.error(\"Could not start #{__MODULE__} because of \" <>\n              \"validation failure: #{inspect(reason)}\")\n\n            :ignore\n        end\n      end\n\n      ..\n    end\n\nThe configuration could be stored in application config:\n\n    # config/ .exs\n    config :my_app, MySocketClient,\n      uri: \"ws://example.org/socket/websocket\",\n      reconnect_after_msec: [200, 500, 1_000, 2_000]\n\nAnd in cases where the configuration validation fails, the `MySocketClient`\nprocess will not crash the application's supervision tree.","ref":"Slipstream.html#c:init/1"},{"type":"callback","title":"Examples - Slipstream.init/1","doc":"@impl Slipstream\n    def init(config) do\n      connect(config)\n    end","ref":"Slipstream.html#c:init/1-examples"},{"type":"function","title":"Slipstream.join/3","doc":"Requests that a topic be joined in the current connection\n\nMultiple topics may be joined by one Slipstream client, but each topic\nmay only be joined once. Despite this, `join/3` may be called on the same\ntopic multiple times, but the result will be idempotent. The client will\nnot request to join unless it has not yet joined that topic. In cases where\nyou wish to begin a new session with a topic, you must first `leave/2` and\nthen `join/3` again.\n\nThe request to join will not error-out if the client is not connected to a\nremote server. In that case, the `join/3` function will act as a no-op.\n\nA join can be awaited in a blocking fashion with `await_join/3`.","ref":"Slipstream.html#join/3"},{"type":"function","title":"Examples - Slipstream.join/3","doc":"@impl Slipstream\n    def handle_connect(socket) do\n      {:ok, join(socket, \"rooms:lobby\", %{user: 1})}\n    end","ref":"Slipstream.html#join/3-examples"},{"type":"function","title":"Slipstream.leave/2","doc":"Requests that the given topic be left\n\nNote that like joining, leaving is an asynchronous request and can be awaited\nwith `await_leave/3`.\n\nAlso similar to `join/3`, `leave/2` is idempotent and will not raise an error\nif the provided topic is not currently joined.","ref":"Slipstream.html#leave/2"},{"type":"function","title":"Examples - Slipstream.leave/2","doc":"iex> {:ok, socket} = leave(socket, \"room:lobby\") |> await_leave(\"rooms:lobby\")\n    iex> join(socket, \"rooms:specific\")","ref":"Slipstream.html#leave/2-examples"},{"type":"function","title":"Slipstream.new_socket/0","doc":"Creates a new socket without immediately connecting to a remote websocket\n\nThis can be useful if you do not wish to request connection with `connect/2`\nduring the `c:init/1` callback (because the `c:init/1` callback requires that\nyou return a `t:Slipstream.Socket.t/0`).","ref":"Slipstream.html#new_socket/0"},{"type":"function","title":"Examples - Slipstream.new_socket/0","doc":"defmodule MySocketClient do\n      use Slipstream\n\n      ..\n\n      @impl Slipstream\n      def init(args) do\n        {:ok, new_socket() |> assign(:init_args, args)}\n      end\n\n      ..\n    end\n\n    iex> new_socket()\n    #Slipstream.Socket","ref":"Slipstream.html#new_socket/0-examples"},{"type":"function","title":"Slipstream.push/5","doc":"Requests that a message be pushed on the websocket connection\n\nA channel has the ability to reply directly to a message, but this reply\nis asynchronous. Handle replies using the `c:handle_reply/3`\ncallback or by awaiting them synchronously with `await_reply/2`.\n\nAlthough this request to the remote server is asynchronous, the call to the\ntransport process to transmit the push is synchronous and will exert\nback-pressure on calls to `push/4`, as `push/4` blocks until the message\nhas been written to the network socket by the transport.\n\nAn `{:ok, ref}` return from `push/4` does not mean the message has been\ndelivered to the phoenix channel or even that it has been sent on the\nnetwork; it only means that the connection was alive as of the most recent\nheartbeat and that the given data was therefore written to the network socket\nusing `:gen_tcp.send/2` or `:ssl.send/2`.\n\nThe only way to confirm that a message was delivered to the phoenix channel\nis if the channel replies to the message. In order to link push requests to\ntheir replies, store the `ref` string returned from the call to `push/4` and\nmatch on it in `c:handle_reply/3`.\n\nIf you are pushing especially large messages, you may need to adjust the\n`timeout` argument so that the GenServer call does not exit with `:timeout`.\nThe default value is `5_000` msec (5 seconds).","ref":"Slipstream.html#push/5"},{"type":"function","title":"Examples - Slipstream.push/5","doc":"@impl Slipstream\n    def handle_join(:success, _response, state) do\n      {:ok, hello_request} = push(socket, \"rooms:lobby\", \"new:msg\", %{user: 1, body: \"hello\"})\n\n      {:ok, Map.put(state, :hello_request, hello_request)}\n    end\n\n    @impl Slipstream\n    def handle_reply(ref, reply, %{hello_request: ref} = state) do\n      IO.inspect(reply, label: \"nice, a response.\")\n\n      {:ok, state}\n    end","ref":"Slipstream.html#push/5-examples"},{"type":"function","title":"Slipstream.push!/4","doc":"Pushes, raising if the topic is not joined or if the channel is dead\n\nSame as `push/4`, but raises in cases of failure.\n\nThis can be useful for pipeing into `await_reply/2`","ref":"Slipstream.html#push!/4"},{"type":"function","title":"Examples - Slipstream.push!/4","doc":"iex> {:ok, result} = push!(socket, \"rooms:lobby\", \"msg:new\", params) |> await_reply()\n    {:ok, %{\"created?\" => true}}","ref":"Slipstream.html#push!/4-examples"},{"type":"function","title":"Slipstream.reconnect/1","doc":"Request reconnection given the last-used connection configuration\n\nNote that when `reconnect/1` is used to re-connect instead of `connect/2`\n(or `connect!/2`), the slipstream process will attempt to reconnect with\na retry mechanism with backoff. The process will wait an interval between\nreconnection attempts following the list of milliseconds provided in the\n`:reconnect_after_msec` key of configuration passed to `connect/2` (or\n`connect!/2`).\n\nThe `c:handle_disconnect/2` callback will be invoked for each\nfailure to re-connect, however, so an implementation of that callback which\nwill simply retry with backoff can be achieved like so:\n\n    @impl Slipstream\n    def handle_disconnect(_reason, socket) do\n      case reconnect(socket) do\n        {:ok, socket} -> {:ok, socket}\n        {:error, reason} -> {:stop, reason, socket}\n      end\n    end\n\n`reconnect/1` may return an `:error` tuple in the case that the socket passed\ndoes not contain any connection information (which is added to the socket\nwith `connect/2` or `connect!/2`), or if the socket is currently connected.\nFor a `reconnect/1` call without configuration, the return pattern is\n`{:error, :no_config}`, and for a socket that is already connected, the\npattern is `{:error, :connected}`.\n\nA reconnect may be awaited with `await_connect/2`.","ref":"Slipstream.html#reconnect/1"},{"type":"function","title":"Slipstream.rejoin/3","doc":"Requests that the specified topic be joined again\n\nIf `params` is not provided, the previously used value will be sent.\n\nIn the case that the specified topic has not been joined before, `rejoin/3`\nwill return `{:error, :never_joined}`.\n\nNote that a rejoin may be awaited with `await_join/3`.","ref":"Slipstream.html#rejoin/3"},{"type":"function","title":"Dealing with crashes - Slipstream.rejoin/3","doc":"When attempting to re-join a disconnected topic with `rejoin/3`, the\nSlipstream process will attempt to use backoff governed by the\n`:rejoin_after_msec` list in configuration passed to `connect/2` (or\n`connect!/2`).\n\nThe `c:handle_topic_close/3` callback will be invoked with\nthe for each crash, however, so a minimal implementation of that callback\nwhich achieves the backoff retry is like so:","ref":"Slipstream.html#rejoin/3-dealing-with-crashes"},{"type":"function","title":"Examples - Slipstream.rejoin/3","doc":"@impl Slipstream\n    def handle_topic_close(topic, _reason, socket) do\n      {:ok, _socket} = rejoin(socket, topic)\n    end","ref":"Slipstream.html#rejoin/3-examples"},{"type":"function","title":"Slipstream.start_link/3","doc":"Starts a slipstream client process\n\nThis function delegates to `GenServer.start_link/3`, so all options passable\nto that function are passable here as well. Most notably, you may name your\nslipstream clients using the same naming rules as GenServers, as in the\nexample below.","ref":"Slipstream.html#start_link/3"},{"type":"function","title":"Examples - Slipstream.start_link/3","doc":"defmodule MySlipstreamClient do\n      use Slipstream\n\n      def start_link(args) do\n        Slipstream.start_link(__MODULE__, args, name: __MODULE__)\n      end\n\n      ..\n    end","ref":"Slipstream.html#start_link/3-examples"},{"type":"callback","title":"Slipstream.terminate/2","doc":"Invoked when a slipstream process is terminated\n\nNote that this callback is not always invoked as the process shuts down.\nSee `c:GenServer.terminate/2` for more information.\n\nIt is wise to `disconnect/1` in this callback (and such is the default\nimplementation). This will gracefully end the websocket connection.\nThis is the behavior of the default implementation of `c:terminate/2`.","ref":"Slipstream.html#c:terminate/2"},{"type":"callback","title":"Examples - Slipstream.terminate/2","doc":"@impl Slipstream\n    def terminate(reason, socket) do\n      Logger.debug(\"shutting down: \" <> inspect(reason))\n\n      disconnect(socket)\n    end","ref":"Slipstream.html#c:terminate/2-examples"},{"type":"type","title":"Slipstream.json_serializable/0","doc":"Any data structure capable of being serialized as JSON\n\nAny argument typed as `t:Slipstream.json_serializable/0` must be able to\nbe encoded with the JSON parser passed in configuration. See\n`Slipstream.Configuration`.","ref":"Slipstream.html#t:json_serializable/0"},{"type":"type","title":"Slipstream.push_reference/0","doc":"A reference to a message pushed by the client\n\nThese references are returned by calls to `push/4` and may be matched on\nin `c:handle_reply/3`. They are also used to match messages\nfor `await_reply/2`.","ref":"Slipstream.html#t:push_reference/0"},{"type":"type","title":"Synchronicity - Slipstream.push_reference/0","doc":"This approach treats the websocket connection as an RPC: some other process\nin the service does a `GenServer.call/3` to the slipstream client process,\nwhich sends a push to the remote websocket server, waits for a reply\n(synchronously) and then sends that back to the caller. All-in-all, this\nappears completely synchronous for the caller.\n\n    @impl Slipstream\n    def handle_call({:new_message, params}, _from, socket) do\n      {:ok, ref} = push(socket, \"rooms:lobby\", \"msg:new\", params)\n\n      {:reply, await_reply(ref), socket}\n    end\n\nThis approach is written in a more asynchronous fashion. An info message\narriving from any other process triggers the slipstream client to push a\nwork message to the remote websocket server. When the remote websocket server\nreplies with the result, the slipstream client sends off the result to be\ndealt with else-where. No process in this scenario blocks, so they are all\ncapable of receiving other messages while the work is being completed.\n\n    @impl Slipstream\n    def handle_info(:do_work, socket) do\n      ref = push(socket, \"worker_queue:foo\", \"do_work\", %{})\n\n      {:noreply, assign(socket, :work_ref, ref)}\n    end\n\n    @impl Slipstream\n    def handle_reply(ref, result, %{assigns: %{work_ref: ref}} = socket) do\n      IO.inspect(result, label: \"work complete!\")\n\n      {:ok, socket}\n    end","ref":"Slipstream.html#t:push_reference/0-synchronicity"},{"type":"type","title":"Slipstream.reply/0","doc":"A reply from a remote server to a push from the client\n\nReplies may be any of\n\n- `:ok`\n- `:error`\n- `{:ok, any()}`\n- `{:error, any()}`\n\ndepending on how the remote server's reply is written.\n\nNote that the empty map is removed in ok and error tuples, so a reply written\nlike so on the server-side:\n\n    def handle_in(_event, _params, socket) do\n      {:reply, {:ok, %{}}, socket}\n    end\n\nwill translate to a reply of `:ok` (and the same for `{:error, %{}}`).","ref":"Slipstream.html#t:reply/0"},{"type":"type","title":"Examples - Slipstream.reply/0","doc":"# on the Phoenix.Channel (server) side:\n    def handle_in(_event, _params, socket) do\n      {:reply, {:ok, %{created?: true}}, socket}\n    end\n\n    # on the Slipstream (client) side:\n    def handle_reply(_ref, {:ok, %{\"created?\" => true}} = _reply, socket) do\n      ..","ref":"Slipstream.html#t:reply/0-examples"},{"type":"module","title":"Slipstream.Configuration","doc":"Configuration for a Slipstream websocket connection\n\nSlipstream server process configuration is passed in with\n`Slipstream.connect/2` (or `Slipstream.connect!/2`), and so all configuration\nis evauated and validated at runtime, as opposed to compile-time validation.\nYou should not expect to see validation errors on configuration unless you\nforce the validation at compile-time, e.g.:\n\n    # you probably don't want to do this...\n    defmodule MyClient do\n      @config Application.compile_env!(:my_app, __MODULE__)\n\n      use Slipstream\n\n      def start_link(args) do\n        Slipstream.start_link(__MODULE__, args, name: __MODULE__)\n      end\n\n      def init(_args), do: {:ok, connect!(@config)}\n\n      ..\n    end\n\nThis approach will validate the configuration at compile-time, but you\nwill be unable to change the configuration after compilation, so any\nsecrets contained in the configuration (e.g. a basic-auth request header)\nwill be compiled into the beam files.\n\nSee the docs for `c:Slipstream.init/1` for a safer approach.","ref":"Slipstream.Configuration.html"},{"type":"module","title":"Options - Slipstream.Configuration","doc":"* `:uri` - Required. The endpoint to which the websocket will connect. Schemes of \"ws\" and\n  \"wss\" are supported, and a scheme must be provided. Either binaries or\n  `URI` structs are accepted. E.g. `\"ws://localhost:4000/socket/websocket\"`.\n\n* `:heartbeat_interval_msec` (`t:non_neg_integer/0`) - The time between heartbeat messages. A value of `0` will disable automatic\n  heartbeat sending. Note that a Phoenix.Channel will close out a connection\n  after 60 seconds of inactivity (`60_000`). The default value is `30000`.\n\n* `:headers` - A set of headers to merge with the request headers when GETing the\n  websocket URI. Headers must be provided as two-tuples where both elements\n  are binaries. Casing of these headers is inconsequential. The default value is `[]`.\n\n* `:serializer` (`t:atom/0`) - A serializer module which exports at least `encode!/1` and `decode!/2`. The default value is `Slipstream.Serializer.PhoenixSocketV2Serializer`.\n\n* `:json_parser` (`t:atom/0`) - A JSON parser module which exports at least `encode!/1` and `decode!/1`. The default value is `Jason`.\n\n* `:reconnect_after_msec` (list of `t:non_neg_integer/0`) - A list of times to reference for trying reconnection when\n  `Slipstream.reconnect/1` is used to request reconnection. The msec time\n  will be fetched based on its position in the list with\n  `Enum.at(reconnect_after_msec, try_number)`. If the number of tries\n  exceeds the length of the list, the final value will be repeated. The default value is `[10, 50, 100, 150, 200, 250, 500, 1000, 2000, 5000]`.\n\n* `:rejoin_after_msec` (list of `t:non_neg_integer/0`) - A list of times to reference for trying to rejoin a topic when\n  `Slipstream.rejoin/3` is used. The msec time\n  will be fetched based on its position in the list with\n  `Enum.at(rejoin_after_msec, try_number)`. If the number of tries\n  exceeds the length of the list, the final value will be repeated. The default value is `[100, 500, 1000, 2000, 5000, 10000]`.\n\n* `:mint_opts` (`t:keyword/0`) - A keywordlist of options to pass to `Mint.HTTP.connect/4` when opening\n  connections. This can be used to set up custom TLS certificate\n  configuration. See the `Mint.HTTP.connect/4` documentation for available\n  options. The default value is `[protocols: [:http1]]`.\n\n* `:extensions` (`t:term/0`) - A list of extensions to pass to `Mint.WebSocket.upgrade/4`. The default value is `[]`.\n\n* `:test_mode?` (`t:boolean/0`) - Whether or not to start-up the client in test-mode. See\n  `Slipstream.SocketTest` for notes on testing Slipstream clients. The default value is `false`.\n\n\n\nNote that a Phoenix.Channel defined with\n\n```elixir\nsocket \"/socket\", UserSocket, ..\n```\n\nCan be connected to at `/socket/websocket`.","ref":"Slipstream.Configuration.html#module-options"},{"type":"function","title":"Slipstream.Configuration.validate/1","doc":"Validates a proposed configuration","ref":"Slipstream.Configuration.html#validate/1"},{"type":"function","title":"Slipstream.Configuration.validate!/1","doc":"Validates a proposed configuration, raising on error","ref":"Slipstream.Configuration.html#validate!/1"},{"type":"type","title":"Slipstream.Configuration.t/0","doc":"","ref":"Slipstream.Configuration.html#t:t/0"},{"type":"module","title":"Slipstream.ConnectionSupervisor","doc":"A supervisor for connection processes\n\nSlipstream.ConnectionSupervisor is a simple module-based `DynamicSupervisor` which\nis used to supervise connection processes. As such, you may track the number\nof connection processes with `DynamicSupervisor.count_children/1`.\nSlipstream.ConnectionSupervisor uses its module name as the DynamicSupervisor name,\nso you may pass `Slipstream.ConnectionSupervisor` into any DynamicSupervisor function.","ref":"Slipstream.ConnectionSupervisor.html"},{"type":"module","title":"Examples - Slipstream.ConnectionSupervisor","doc":"iex> Slipstream.ConnectionSupervisor |> DynamicSupervisor.count_children()\n    %{active: 15, specs: 15, supervisors: 0, workers: 15}","ref":"Slipstream.ConnectionSupervisor.html#module-examples"},{"type":"function","title":"Slipstream.ConnectionSupervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Slipstream.ConnectionSupervisor.html#child_spec/1"},{"type":"behaviour","title":"Slipstream.Serializer","doc":"A behaviour that serializes incoming and outgoing socket messages.","ref":"Slipstream.Serializer.html"},{"type":"callback","title":"Slipstream.Serializer.decode!/2","doc":"Decodes binary into `Slipstream.Message` struct.","ref":"Slipstream.Serializer.html#c:decode!/2"},{"type":"callback","title":"Slipstream.Serializer.encode!/2","doc":"Encodes `Slipstream.Message` structs to binary.","ref":"Slipstream.Serializer.html#c:encode!/2"},{"type":"module","title":"Slipstream.Serializer.PhoenixSocketV2Serializer","doc":"A client-side implementation that corresponds to the server-side Phoenix.Socket.V2.JSONSerializer.","ref":"Slipstream.Serializer.PhoenixSocketV2Serializer.html"},{"type":"function","title":"Slipstream.Serializer.PhoenixSocketV2Serializer.decode!/2","doc":"","ref":"Slipstream.Serializer.PhoenixSocketV2Serializer.html#decode!/2"},{"type":"function","title":"Slipstream.Serializer.PhoenixSocketV2Serializer.encode!/2","doc":"","ref":"Slipstream.Serializer.PhoenixSocketV2Serializer.html#encode!/2"},{"type":"module","title":"Slipstream.Socket","doc":"A data structure representing a potential websocket client connection\n\nThis structure closely resembles `t:Phoenix.Socket.t/0`, but is not\ncompatible with its functions. All documented functions from this module\nare imported by `use Slipstream`.","ref":"Slipstream.Socket.html"},{"type":"function","title":"Slipstream.Socket.assign/2","doc":"","ref":"Slipstream.Socket.html#assign/2"},{"type":"function","title":"Slipstream.Socket.assign/3","doc":"Adds key-value pairs to socket assigns\n\nBehaves the same as `Phoenix.Socket.assign/3`","ref":"Slipstream.Socket.html#assign/3"},{"type":"function","title":"Examples - Slipstream.Socket.assign/3","doc":"iex> assign(socket, :key, :value)\n    iex> assign(socket, key: :value)","ref":"Slipstream.Socket.html#assign/3-examples"},{"type":"function","title":"Slipstream.Socket.channel_pid/1","doc":"Gets the process ID of the connection\n\nThe slipstream implementor module is not the same process as the GenServer\nwhich interfaces with the remote server for websocket communication. This\nother process, the Slipstream.Connection process, interfaces with the\nlow-level WebSocket connection and communicates with the implementor module\nby puassing messages (mostly with `Kernel.send/2`).\n\nIt can be useful to have access to this pid for testing or debugging\npurposes, such as sending a fake disconnect message or for getting state\nwith `:sys.get_state/1`.","ref":"Slipstream.Socket.html#channel_pid/1"},{"type":"function","title":"Examples - Slipstream.Socket.channel_pid/1","doc":"iex> Slipstream.Socket.channel_pid(socket)\n    #PID<0.1.2>","ref":"Slipstream.Socket.html#channel_pid/1-examples"},{"type":"function","title":"Slipstream.Socket.connected?/1","doc":"Checks if a socket is connected to a remote websocket host","ref":"Slipstream.Socket.html#connected?/1"},{"type":"function","title":"Examples - Slipstream.Socket.connected?/1","doc":"iex> socket = connect(socket, uri: \"ws://example.org\")\n    iex> socket = await_connect!(socket)\n    iex> connected?(socket)\n    true","ref":"Slipstream.Socket.html#connected?/1-examples"},{"type":"function","title":"Slipstream.Socket.join_status/2","doc":"Checks the status of a join request\n\nWhen a join is requested with `Slipstream.join/3`, the join request is\nconsidered to be in the `:requested` state. Once the topic is successfully\njoined, it is considered `:joined` until closed. If there is a failure to\njoin the topic, if the topic crashes, or if the topic is left after being\njoined, the status of the join is considered `:closed`. Finally, if a topic\nhas not been requested in a join so far for a socket, the status is `nil`.\n\nNotably, the status of a join will not automatically change to `:joined` once\nthe remote server replies with successful join. Either the join must be\nawaited with `Slipstream.await_join/2` or the status may be checked later\nin the `c:Slipstream.handle_join/3` callback.","ref":"Slipstream.Socket.html#join_status/2"},{"type":"function","title":"Examples - Slipstream.Socket.join_status/2","doc":"iex> socket = join(socket, \"room:lobby\")\n    iex> join_status(socket, \"room:lobby\")\n    :requested\n    iex> {:ok, socket, _join_response} = await_join(socket, \"room:lobby\")\n    iex> join_status(socket, \"room:lobby\")\n    :joined","ref":"Slipstream.Socket.html#join_status/2-examples"},{"type":"function","title":"Slipstream.Socket.joined?/2","doc":"Checks if a channel is currently joined","ref":"Slipstream.Socket.html#joined?/2"},{"type":"function","title":"Examples - Slipstream.Socket.joined?/2","doc":"iex> joined?(socket, \"room:lobby\")\n    true","ref":"Slipstream.Socket.html#joined?/2-examples"},{"type":"function","title":"Slipstream.Socket.update/3","doc":"Updates an existing key in the socket assigns\n\nRaises a `KeyError` if the key is not present in `socket.assigns`.\n\n`func` should be an 1-arity function which takes the existing value at assign\n`key` and updates it to a new value. The new value will take the old value's\nplace in `socket.assigns[key]`.\n\nThis function is a useful alternative to `assign/3` when the key is already\npresent in assigns and is a list, map, or similarly malleable data structure.","ref":"Slipstream.Socket.html#update/3"},{"type":"function","title":"Examples - Slipstream.Socket.update/3","doc":"@impl Slipstream\n    def handle_cast({:join, topic}, socket) do\n      socket =\n        socket\n        |> update(:topics, &[topic | &1])\n        |> join(topic)\n\n      {:noreply, socket}\n    end\n\n    @impl Slipstream\n    def handle_call({:join, topic}, from, socket) do\n      socket =\n        socket\n        |> update(:join_requests, &Map.put(&1, topic, from))\n        |> join(topic)\n\n      # note: not replying here so we can provide a synchronous call to a\n      # topic being joined\n      {:noreply, socket}\n    end\n\n    @impl Slipstream\n    def handle_join(topic, response, socket) do\n      case Map.fetch(socket.assigns.join_requests, topic) do\n        {:ok, from} -> GenServer.reply(from, {:ok, response})\n        :error -> :ok\n      end\n\n      {:ok, socket}\n    end","ref":"Slipstream.Socket.html#update/3-examples"},{"type":"type","title":"Slipstream.Socket.t/0","doc":"A socket data structure representing a potential websocket client connection","ref":"Slipstream.Socket.html#t:t/0"},{"type":"module","title":"Slipstream.SocketTest","doc":"Helper functions and macros for testing Slipstream clients\n\nThis module is something of a correlary to `Phoenix.ChannelTest`. The\nfunctions and macros in `Phoenix.ChannelTest` emulate client operations:\nfunctions like `Phoenix.ChannelTest.join/2`, `Phoenix.ChannelTest.push/3`,\netc.. Functions and macros in this module emulate behavior of the server:\na `Phoenix.Channel`.","ref":"Slipstream.SocketTest.html"},{"type":"module","title":"Timing assumptions - Slipstream.SocketTest","doc":"Clients are typically written to assume that the server\n\n- is up at time of client start-up\n- is always up\n\nWhile this is typically (at least mostly) accurate, it is not necessarily\ntrue in general and will not be true when testing Slipstream clients. In\nparticular, Slipstream clients may fail in test-mode under the following\nconditions:\n\n- the client is started immediately in the application supervision tree\n- it awaits a connection synchronously (with `Slipstream.await_connect/2`)\n- the testing suite takes longer to complete than the timeout given to\n  `Slipstream.await_connect/2`\n\nWhile a client that uses `Slipstream.await_connect/2` can band-aid over\nthese problems with a high enough timeout, clients should be satisfied with\nreceiving successful connection events asynchronously. Clients written in\nthe asynchronous callback style will not be affected.\n\nSome implementation-level details are glossed over by this testing framework,\nincluding that of heartbeat timeouts. If a client sits idle after joining\nfor more than 60 seconds, they will not be terminated due to heartbeat\ntimeout.","ref":"Slipstream.SocketTest.html#module-timing-assumptions"},{"type":"module","title":"Setting up a test - Slipstream.SocketTest","doc":"Another assumption clients typically make of servers is that clients assume\nthere is just _one_ server. A client is not written expecting to hear that it\nhas been connected multiple times for one request to `Slipstream.connect/2`.\nAs such, tests using this module as a case template should be run\nsynchronously.\n\n    defmodule MyApp.MyClientTest do\n      use Slipstream.SocketTest\n\n      ..\n\nBy default, this will start a server session for each test that simulates\nthe current test process as the websocket server connected to the client.\n\n    test \"the client sends a push to the server on join\", c do\n      accept_connect(MyClient)\n    end\n\nThis server does not run a websocket server. Instead the server is\na conceptual server: you may imagine that in each test, you are have control\nof the remote server and can control the behavior of the server imperatively.\n\nThe `assert_*` and `refute_*` family of macros from this module allow you to\nmake assertions about- and match on values from- requests from the client to\nthe server. The remaining functions allow you to emulate actions on behalf\nof a hypothetical server.","ref":"Slipstream.SocketTest.html#module-setting-up-a-test"},{"type":"module","title":"Starting the client - Slipstream.SocketTest","doc":"When working with `accept_conect/1` and `connect_and_assert_join/5`, you may\neither pass the name of the GenServer to test, or a pid. If starting the client\nyourself, make sure that you pass the `test_mode?: true` option, otherwise the\nSlipstream client will attempt to connect to the configured uri.\n\n    defmodule MyApp.MyClientTest do\n      use Slipstream.SocketTest\n\n      setup do\n        client = start_supervised!(MyApp.Myclient, uri: \"wss://test.com\", test_mode?: true)\n        %{client: client}\n      end","ref":"Slipstream.SocketTest.html#module-starting-the-client"},{"type":"module","title":"Timeouts - Slipstream.SocketTest","doc":"The `assert_*` and `refute_*` macros from this module default to ExUnit\ntimeouts. See the ExUnit documentation for more details.","ref":"Slipstream.SocketTest.html#module-timeouts"},{"type":"module","title":"Formatting - Slipstream.SocketTest","doc":"Slipstream exports the `assert_*` and `receive_*` macros as valid\n`locals_without_parens` in its formatter config. To be able to use these\nmacros without the formatter injecting parentheses, import `:slipstream` in\nyour service's formatter config for `:import_deps`:\n\n    [\n      inputs: [\"mix.exs\", \"{config,lib,test}/**/*.{ex,exs}\"],\n      import_deps: [:slipstream],\n      .. # more configuration\n    ]","ref":"Slipstream.SocketTest.html#module-formatting"},{"type":"function","title":"Slipstream.SocketTest.accept_connect/1","doc":"Emulates a server telling the client it has connected\n\nThis sets the current process as the current server connected to the client.\nIt also adds an `ExUnit.callbacks.on_exit/2` function that disconnects the\nclient on exit with reason `:closed_by_test`. To handle this disconnect,\nclients should match on this pattern in `c:Slipstream.handle_disconnect/2`\nand either shutdown (in the case that the test controls the spawning of the\nclient, or reconnect, as with `Slipstream.reconnect/1`. The default\nimplementation of `c:Slipstream.handle_disconnect/2` reconnects in this case.\n\nSee [Timing Assumptions](#module-timing-assumptions).","ref":"Slipstream.SocketTest.html#accept_connect/1"},{"type":"function","title":"Examples - Slipstream.SocketTest.accept_connect/1","doc":"test \"the server connects\", c do\n      :ok = accept_connect(MyApp.MyClient)\n    end","ref":"Slipstream.SocketTest.html#accept_connect/1-examples"},{"type":"macro","title":"Slipstream.SocketTest.assert_disconnect/1","doc":"Asserts that a client will attempt to disconnect from the server","ref":"Slipstream.SocketTest.html#assert_disconnect/1"},{"type":"macro","title":"Examples - Slipstream.SocketTest.assert_disconnect/1","doc":"accept_connect(MyClient)\n    # client will disconnect after 15s of inactivity\n    assert_disconnect 15_000","ref":"Slipstream.SocketTest.html#assert_disconnect/1-examples"},{"type":"macro","title":"Slipstream.SocketTest.assert_join/4","doc":"Asserts that a client will request to join a topic\n\n`topic_expr` and `params_expr` are interpreted as match expressions, so they\nmay be literal values, pinned (`^`) bindings, or partial values such as\n`\"msg:\" <> _` or `%{}` (which matches any map).\n\n`reply` is meant to simulate the return value of the\n`c:Phoenix.Channel.join/3` callback.","ref":"Slipstream.SocketTest.html#assert_join/4"},{"type":"macro","title":"Examples - Slipstream.SocketTest.assert_join/4","doc":"accept_connect(MyClient)\n    assert_join \"rooms:lobby\", %{}, :ok","ref":"Slipstream.SocketTest.html#assert_join/4-examples"},{"type":"macro","title":"Slipstream.SocketTest.assert_leave/2","doc":"Asserts that a client will request to leave a topic\n\n`topic_expr` is a pattern, so literal values like `\"room:lobby\"` are valid\nas well as match patterns such as `\"room:\" <> _`. Existing bindings must be\npinned with the `^` pin operator.","ref":"Slipstream.SocketTest.html#assert_leave/2"},{"type":"macro","title":"Examples - Slipstream.SocketTest.assert_leave/2","doc":"topic = \"rooms:lobby\"\n    accept_connect(MyClient)\n    assert_join ^topic, %{}, :ok\n    push(MyClient, topic, \"leave\", %{})\n    assert_leave ^topic","ref":"Slipstream.SocketTest.html#assert_leave/2-examples"},{"type":"macro","title":"Slipstream.SocketTest.assert_push/5","doc":"Asserts that the client will request to push a message to the server\n\nNote that `topic_expr`, `event_expr`, `params_expr`, and `ref_expr` are all\npattern expressions. Prior bindings may be used with the `^` pin operator,\nvalues may be underscored to ignore, and partial values may be matched (e.g.\n`%{}` will match any map).\n\n`ref_expr` can be provided to bind a reference for later use in `reply/3`.","ref":"Slipstream.SocketTest.html#assert_push/5"},{"type":"macro","title":"Examples - Slipstream.SocketTest.assert_push/5","doc":"assert_push \"rooms:lobby\", \"msg:new\", params, ref\n    reply(MyClient, ref, {:ok, %{status: \"ok\", received: params}})","ref":"Slipstream.SocketTest.html#assert_push/5-examples"},{"type":"macro","title":"Slipstream.SocketTest.connect_and_assert_join/5","doc":"A convenience macro wrapping connection and a join response\n\nThis macro is written for clients that join immediately after a connection has\nbeen established, which is a common case.\n\nClients written like so:\n\n    @impl Slipstream\n    def handle_connect(socket) do\n      {:ok, join(socket, \"rooms:lobby\", %{user_id: socket.assigns.user_id})}\n    end\n\nMay be tested with `connect_and_assert_join/5` as opposed to a separate `connect/2`\nand then an `assert_join/5`.","ref":"Slipstream.SocketTest.html#connect_and_assert_join/5"},{"type":"macro","title":"Examples - Slipstream.SocketTest.connect_and_assert_join/5","doc":"socket = connect_and_assert_join MySocketClient, \"rooms:lobby\", %{}, :ok\n    push(socket, \"rooms:lobby\", \"initial-hello\", %{\"hello\" => \"world\"})","ref":"Slipstream.SocketTest.html#connect_and_assert_join/5-examples"},{"type":"function","title":"Slipstream.SocketTest.disconnect/2","doc":"Emulates a server closing a connection to the client","ref":"Slipstream.SocketTest.html#disconnect/2"},{"type":"function","title":"Examples - Slipstream.SocketTest.disconnect/2","doc":"accept_connect(MyClient)\n    disconnect(MyClient, :heartbeat_timeout)","ref":"Slipstream.SocketTest.html#disconnect/2-examples"},{"type":"function","title":"Slipstream.SocketTest.push/4","doc":"Emulates a server pushing a message to the client\n\nThis emulates cases of `Phoenix.Channel.push/3` by a Phoenix server.\n\nNote that this function will not encode or decode the value of `params`, so\nconversion from maps of atom-keys to string-keys will not occur as it would\nwhen passing messages over-the-wire.","ref":"Slipstream.SocketTest.html#push/4"},{"type":"function","title":"Examples - Slipstream.SocketTest.push/4","doc":"test \"the server increments our counter with ping messages\", c do\n      assert Counter.count() == 0\n\n      connect_and_assert_join MyClient, \"counter-topic\", %{}, :ok\n\n      push(MyClient, \"counter-topic\", \"ping\", %{delta: 1})\n\n      assert Counter.count() == 1\n    end","ref":"Slipstream.SocketTest.html#push/4-examples"},{"type":"macro","title":"Slipstream.SocketTest.refute_disconnect/1","doc":"Refutes that a client will attempt to disconnect from the server\n\nThe opposite of `assert_disconnect/1`.","ref":"Slipstream.SocketTest.html#refute_disconnect/1"},{"type":"macro","title":"Examples - Slipstream.SocketTest.refute_disconnect/1","doc":"accept_connect(MyClient)\n    refute_disconnect 10_000","ref":"Slipstream.SocketTest.html#refute_disconnect/1-examples"},{"type":"macro","title":"Slipstream.SocketTest.refute_join/3","doc":"Refutes that a client will request to join a topic\n\nThe opposite of `assert_join/5`.","ref":"Slipstream.SocketTest.html#refute_join/3"},{"type":"macro","title":"Examples - Slipstream.SocketTest.refute_join/3","doc":"accept_connect(MySocketClient)\n    refute_join \"rooms:\" <> _, %{user_id: 5}","ref":"Slipstream.SocketTest.html#refute_join/3-examples"},{"type":"macro","title":"Slipstream.SocketTest.refute_leave/2","doc":"Refutes that a client will request to leave a topic\n\nThe opposite of `assert_leave/3`.","ref":"Slipstream.SocketTest.html#refute_leave/2"},{"type":"macro","title":"Examples - Slipstream.SocketTest.refute_leave/2","doc":"accept_connect(MyClient)\n    assert_join \"rooms:lobby\", %{}, :ok\n    push(MyClient, topic, \"no-don't-go\", %{})\n    refute_leave ^topic, 10_000","ref":"Slipstream.SocketTest.html#refute_leave/2-examples"},{"type":"macro","title":"Slipstream.SocketTest.refute_push/4","doc":"Refutes that a client will push a message to the server\n\nThe opposite of `assert_push/4`","ref":"Slipstream.SocketTest.html#refute_push/4"},{"type":"macro","title":"Examples - Slipstream.SocketTest.refute_push/4","doc":"refute_push \"rooms:lobby\", \"msg:\" <> _, %{user_id: 5}","ref":"Slipstream.SocketTest.html#refute_push/4-examples"},{"type":"function","title":"Slipstream.SocketTest.reply/3","doc":"Emulates a server replying to a push from the client\n\n`ref` is a reference which can be matched upon in `assert_push/6`. `reply`\nfollows the `t:Slipstream.reply/0` type: it may be an `:ok` or `:error`\natom or an `{:ok, any()}` or `{:error, any()}` tuple. This value will not\nbe encoded or decoded by Slipstream, instead just directly passed to the\nclient.","ref":"Slipstream.SocketTest.html#reply/3"},{"type":"function","title":"Examples - Slipstream.SocketTest.reply/3","doc":"topic = \"rooms:lobby\"\n    connect_and_assert_join MySocketClient, ^topic, %{}, :ok\n    assert_push ^topic, \"ping\", %{}, ref\n    reply(MySocketClient, ref, {:ok, %{\"ping\" => \"pong\"}})","ref":"Slipstream.SocketTest.html#reply/3-examples"},{"type":"type","title":"Slipstream.SocketTest.client/0","doc":"Any Slipstream client\n\nSince Slipstream clients are either GenServer or plain processes, either\na pid or a GenServer name will work as the `client` argument for any function\nspecifying `client` in this module.","ref":"Slipstream.SocketTest.html#t:client/0"},{"type":"exception","title":"Slipstream.Serializer.DecodeError","doc":"","ref":"Slipstream.Serializer.DecodeError.html"},{"type":"type","title":"Slipstream.Serializer.DecodeError.t/0","doc":"","ref":"Slipstream.Serializer.DecodeError.html#t:t/0"},{"type":"exception","title":"Slipstream.Serializer.EncodeError","doc":"","ref":"Slipstream.Serializer.EncodeError.html"},{"type":"type","title":"Slipstream.Serializer.EncodeError.t/0","doc":"","ref":"Slipstream.Serializer.EncodeError.html#t:t/0"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\nThe format is based on [Keep a\nChangelog](https://keepachangelog.com/en/1.0.0/), and this project adheres to\n[Semantic Versioning](https://semver.org/spec/v2.0.0.html).","ref":"changelog.html"},{"type":"extras","title":"1.1.3 - 2024-12-04 - Changelog","doc":"","ref":"changelog.html#1-1-3-2024-12-04"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed callback return types to include `{:noreply, new_socket}` and\n  `{:noreply, new_socket, _rest}` (for returning timeouts or triggering\n  `handle_continue/2` for example). These return types were always\n  supported but not declared in the callback, so the dialyzer would\n  complain if these values were returned.\n    - `c:Slipstream.handle_connect/1`\n    - `c:Slipstream.handle_disconnect/2`\n    - `c:Slipstream.handle_join/3`\n    - `c:Slipstream.handle_message/4`\n    - `c:Slipstream.handle_reply/3`\n    - `c:Slipstream.handle_topic_close/3`\n    - `c:Slipstream.handle_leave/2`","ref":"changelog.html#fixed"},{"type":"extras","title":"1.1.2 - 2024-09-26 - Changelog","doc":"","ref":"changelog.html#1-1-2-2024-09-26"},{"type":"extras","title":"Added - Changelog","doc":"- Clarified and enhanced `Slipstream.push/4` docs\n- Clarified and enhanced `Slipstream` module docs around retries","ref":"changelog.html#added"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed a type error on Elixir 1.17+ about an unknown `.message` key.","ref":"changelog.html#fixed"},{"type":"extras","title":"1.1.1 - 2024-01-26 - Changelog","doc":"","ref":"changelog.html#1-1-1-2024-01-26"},{"type":"extras","title":"Fixed - Changelog","doc":"- Added deserialization clauses to the default (de)serializer for binary\n  broadcast messages.\n    - Broadcasts like `MyAppWeb.Endpoint.broadcast!(topic, event, {:binary, <<1, 2, 3>>})`\n      improperly arrived as a message in `c:Slipstream.handle_info/2` prior to this fix.","ref":"changelog.html#fixed"},{"type":"extras","title":"1.1.0 - 2023-06-12 - Changelog","doc":"","ref":"changelog.html#1-1-0-2023-06-12"},{"type":"extras","title":"Added - Changelog","doc":"- Added configuration for a serializer module with the default module\n  implementing `Phoenix.Socket.V2.JSONSerializer` (Phoenix's default\n  serializer).\n- Added support for sending and receiving binary data.","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- The `:json_parser` key previously required a module that exported\n  `encode!/1` and `decode/1`. This key now requires a module that\n  implements `encode!/1` and `decode!/1`.\n    - This was previously documented as requiring `encode/1` but this\n      was a typo in the documentation.\n    - Note that many popular JSON parser/generator libraries share\n      these function signatures (`encode/1`, `encode!/1`, `decode/1` and\n      `decode!/1`), so it is unlikely that you will need to make any\n      changes in your code.","ref":"changelog.html#changed"},{"type":"extras","title":"1.0.4 - 2023-04-02 - Changelog","doc":"","ref":"changelog.html#1-0-4-2023-04-02"},{"type":"extras","title":"Fixed - Changelog","doc":"- Added the `join_ref` to the `phx_leave` message.\n    - This fixes compatibility with Phoenix 1.6.16+ when using\n      `Slipstream.leave/2`.","ref":"changelog.html#fixed"},{"type":"extras","title":"1.0.3 - 2023-03-06 - Changelog","doc":"","ref":"changelog.html#1-0-3-2023-03-06"},{"type":"extras","title":"Fixed - Changelog","doc":"- `nimble_options` is now allowed at `~> 1.0 or ~> 0.1`.\n    - NimbleOptions' v1.0.0 release contains no breaking changes from the 0.x\n      release series.","ref":"changelog.html#fixed"},{"type":"extras","title":"1.0.2 - 2023-03-06 - Changelog","doc":"","ref":"changelog.html#1-0-2-2023-03-06"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed a memory growth issue that could occur when a socket failed to join a\n  channel repeatedly.\n    - Memory usage of the Slipstream.Connection process and the client process\n      could grow sharply as the socket tried to rejoin the failed topic, with\n      higher memory usage per retry and per unjoined topic.","ref":"changelog.html#fixed"},{"type":"extras","title":"1.0.1 - 2022-06-22 - Changelog","doc":"","ref":"changelog.html#1-0-1-2022-06-22"},{"type":"extras","title":"Fixed - Changelog","doc":"- Any `Mint.TransportError`s received will now close the WebSocket connection","ref":"changelog.html#fixed"},{"type":"extras","title":"1.0.0 - 2022-05-11 - Changelog","doc":"This release represents stability in the API. There are no functional changes\nbetween this release and v0.8.5.","ref":"changelog.html#1-0-0-2022-05-11"},{"type":"extras","title":"0.8.5 - 2022-04-26 - Changelog","doc":"","ref":"changelog.html#0-8-5-2022-04-26"},{"type":"extras","title":"Changed - Changelog","doc":"- Widened compatibility with `mint_web_socket` to `~> 1.0` or `~> 0.2`","ref":"changelog.html#changed"},{"type":"extras","title":"0.8.4 - 2022-02-17 - Changelog","doc":"","ref":"changelog.html#0-8-4-2022-02-17"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed compatibility with `mint_web_socket` v0.2.0","ref":"changelog.html#fixed"},{"type":"extras","title":"0.8.3 - 2021-10-07 - Changelog","doc":"","ref":"changelog.html#0-8-3-2021-10-07"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed connection closing when `Mint.HTTP.stream/2` returns errors\n    - [#42](https://github.com/NFIBrokerage/slipstream/42) thanks [`@parallel588`](https://github.com/parallel588)!","ref":"changelog.html#fixed"},{"type":"extras","title":"0.8.2 - 2021-10-01 - Changelog","doc":"","ref":"changelog.html#0-8-2-2021-10-01"},{"type":"extras","title":"Changed - Changelog","doc":"- Expanded dependency on `:telemetry` to allow `~> 1.0`\n    - [#41](https://github.com/NFIBrokerage/slipstream/pull/41) thanks [`@fhunleth`](https://github.com/fhunleth)!","ref":"changelog.html#changed"},{"type":"extras","title":"0.8.1 - 2021-07-06 - Changelog","doc":"","ref":"changelog.html#0-8-1-2021-07-06"},{"type":"extras","title":"Fixed - Changelog","doc":"- Properly emit the channel-closed event when `Mint.HTTP.stream/2` returns an\n  error tuple about the connection being closed","ref":"changelog.html#fixed"},{"type":"extras","title":"0.8.0 - 2021-07-01 - Changelog","doc":"","ref":"changelog.html#0-8-0-2021-07-01"},{"type":"extras","title":"Changed - Changelog","doc":"- Switched out `:gun` for Mint.WebSocket as the low-level websocket client","ref":"changelog.html#changed"},{"type":"extras","title":"Removed - Changelog","doc":"- Removed the `:gun_open_options` key from configuration\n    - Use the new `:mint_opts` key instead to configure TLS options","ref":"changelog.html#removed"},{"type":"extras","title":"Added - Changelog","doc":"- Added the `:mint_opts` key to configuration for controlling the options\n  passed to `Mint.HTTP.connect/4`\n- Added the `:extensions` key to configuration for specifying\n  `Mint.WebSocket.Extension` configuration passed to `Mint.WebSocket.upgrade/4`","ref":"changelog.html#added"},{"type":"extras","title":"0.7.0 - 2021-06-21 - Changelog","doc":"","ref":"changelog.html#0-7-0-2021-06-21"},{"type":"extras","title":"Added - Changelog","doc":"- Added `c:Slipstream.handle_leave/2`\n    - This is invoked when the server acknowledges that the client has left\n      the topic\n    - This was previously invoked as the more generic\n      `c:Slipstream.handle_topic_close/3`","ref":"changelog.html#added"},{"type":"extras","title":"Changed - Changelog","doc":"- Topic leaves are now handled not in `c:Slipstream.handle_topic_close/3`\n  but `c:Slipstream.handle_leave/2`\n    - **this is a breaking change**\n\nIn order to migrate existing code, change any implementations of\n`c:Slipstream.handle_topic_close/3` with the `reason` of `:left` from\n\n```elixir\ndef handle_topic_close(_topic, :left, socket)\n```\n\nTo\n\n```elixir\ndef handle_leave(_topic, socket)\n```\n\nNote that this callback has a default behavior of performing a no-op.","ref":"changelog.html#changed"},{"type":"extras","title":"Fixed - Changelog","doc":"- Failures to connect (via an `:error` return from `c:Phoenix.Socket.connect/3`)\n  now correctly trigger a `Slipstream.Events.ChannelConnectFailed` event\n    - when using the synchronous API, this will result in an error tuple with\n      `Slipstream.await_connect/2` in the format of\n      `{:error, {:connect_failure, %{resp_headers: resp_headers, status_code: status_code}}}`\n      where `status_code` will be `403`.\n    - when using the module-based API, this will invoke the\n      `c:Slipstream.handle_disconnect/2` callback with the same error tuple","ref":"changelog.html#fixed"},{"type":"extras","title":"0.6.2 - 2021-03-01 - Changelog","doc":"","ref":"changelog.html#0-6-2-2021-03-01"},{"type":"extras","title":"Fixed - Changelog","doc":"- `await_join!/2` and `await_leave!/2` have been changed to `await_join!/3`\n  and `await_leave!/3` matching their intended usage\n    - the `topic` parameter has been added which coincides with the API of\n      `await_join/3` and `await_leave/3`","ref":"changelog.html#fixed"},{"type":"extras","title":"0.6.1 - 2021-02-27 - Changelog","doc":"","ref":"changelog.html#0-6-1-2021-02-27"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed dialyzer warnings arising from some code macro-injected by\n  `Slipstream.__using__/1`\n    - thanks [`@jjcarstens`](https://github.com/jjcarstens)!","ref":"changelog.html#fixed"},{"type":"extras","title":"0.6.0 - 2021-02-24 - Changelog","doc":"","ref":"changelog.html#0-6-0-2021-02-24"},{"type":"extras","title":"Removed - Changelog","doc":"- Removed dependency on Phoenix","ref":"changelog.html#removed"},{"type":"extras","title":"0.5.4 - 2021-02-22 - Changelog","doc":"","ref":"changelog.html#0-5-4-2021-02-22"},{"type":"extras","title":"Fixed - Changelog","doc":"- Slipstream will now gracefully handle failures from `:gun.open/3`\n    - errors will result in an invocation of a client's\n      `c:Slipstream.handle_disconnect/2` callback (if the client is a module)\n      or return an error tuple from `await_connect/2` in the case of a\n      synchronous client","ref":"changelog.html#fixed"},{"type":"extras","title":"0.5.3 - 2021-02-11 - Changelog","doc":"","ref":"changelog.html#0-5-3-2021-02-11"},{"type":"extras","title":"Changed - Changelog","doc":"- Synchronous functions are now grouped under \"Synchronous Functions\" in the\n  documentation","ref":"changelog.html#changed"},{"type":"extras","title":"0.5.2 - 2021-02-10 - Changelog","doc":"","ref":"changelog.html#0-5-2-2021-02-10"},{"type":"extras","title":"Fixed - Changelog","doc":"- The timer reference for the heartbeat interval is now properly matched upon\n  in the connection process\n    - this fixes some behavior where the timer reference would not be canceled\n      (with `:timer.cancel/1`) upon disconnection\n    - this has not manifested itself as a bug as far as we are aware, but this\n      fix should properly clean up the timer when it is no longer needed","ref":"changelog.html#fixed"},{"type":"extras","title":"0.5.1 - 2021-02-08 - Changelog","doc":"","ref":"changelog.html#0-5-1-2021-02-08"},{"type":"extras","title":"Added - Changelog","doc":"- Added an example guide on how to use GenServer operations for a client\n- Added an example on using Slipstream to script interactions with remote\n  `Phoenix.Channel`s","ref":"changelog.html#added"},{"type":"extras","title":"0.5.0 - 2021-02-07 - Changelog","doc":"","ref":"changelog.html#0-5-0-2021-02-07"},{"type":"extras","title":"Added - Changelog","doc":"- Added `Slipstream.Socket.update/3` which emulates `Phoenix.LiveView.update/3`\n- Added an example on the rejoin-after-reconnect pattern for clients which\n  dynamically join topics","ref":"changelog.html#added"},{"type":"extras","title":"0.4.4 - 2021-02-07 - Changelog","doc":"","ref":"changelog.html#0-4-4-2021-02-07"},{"type":"extras","title":"Fixed - Changelog","doc":"- Link to post-cutover-metrics image made absolute","ref":"changelog.html#fixed"},{"type":"extras","title":"0.4.3 - 2021-02-07 - Changelog","doc":"","ref":"changelog.html#0-4-3-2021-02-07"},{"type":"extras","title":"Added - Changelog","doc":"- Added examples to the documentation\n    - \"Graceful Startup\" and \"Repeater-style Client\"\n- Added a note about post-cutover performance changes after we (NFIBrokerage)\n  cutover to Slipstream in our stack\n    - This is a new section in the \"Why :gun?\" guide\n- Added documentation on counting the number of Slipstream connection processes\n  to `Slipstream.ConnectionSupervisor`","ref":"changelog.html#added"},{"type":"extras","title":"0.4.2 - 2021-02-05 - Changelog","doc":"","ref":"changelog.html#0-4-2-2021-02-05"},{"type":"extras","title":"Fixed - Changelog","doc":"- Added channel config to socket on `Slipstream.connect/2` or\n  `Slipstream.connect!/2`\n    - this fixes an issue reconnecting when faking a connection through\n      `Slipstream.SocketTest.accept_connect/1`","ref":"changelog.html#fixed"},{"type":"extras","title":"0.4.1 - 2021-02-04 - Changelog","doc":"","ref":"changelog.html#0-4-1-2021-02-04"},{"type":"extras","title":"Fixed - Changelog","doc":"- Changed the default behavior of `c:Slipstream.terminate/2` to disconnect\n  gracefully with `Slipstream.disconnect/1`","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- Added better documentation on the `c:Slipstream.terminate/2` callback and\n  how to disconnect in it to leave a connection gracefully","ref":"changelog.html#added"},{"type":"extras","title":"0.4.0 - 2021-02-04 - Changelog","doc":"","ref":"changelog.html#0-4-0-2021-02-04"},{"type":"extras","title":"Added - Changelog","doc":"- Added telemetry support for clients\n    - see the telemetry guide for more details","ref":"changelog.html#added"},{"type":"extras","title":"0.3.4 - 2021-02-04 - Changelog","doc":"","ref":"changelog.html#0-3-4-2021-02-04"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed spelling and grammar mistakes in the implementation guide","ref":"changelog.html#fixed"},{"type":"extras","title":"0.3.3 - 2021-02-04 - Changelog","doc":"","ref":"changelog.html#0-3-3-2021-02-04"},{"type":"extras","title":"Added - Changelog","doc":"- increased documentation around retry mechanisms\n- added a page on the choice of `:gun` as the low-level webscket client","ref":"changelog.html#added"},{"type":"extras","title":"0.3.2 - 2021-02-03 - Changelog","doc":"","ref":"changelog.html#0-3-2-2021-02-03"},{"type":"extras","title":"Fixed - Changelog","doc":"- Fixed edge case where the remote server shutting down would not emit a\n  `ChannelClosed` event\n    - this and heartbeats caused the odd `:gun_error` atom from\n      [#12](https://github.com/NFIBrokerage/slipstream/issues/12)","ref":"changelog.html#fixed"},{"type":"extras","title":"0.3.1 - 2021-02-03 - Changelog","doc":"","ref":"changelog.html#0-3-1-2021-02-03"},{"type":"extras","title":"Fixed - Changelog","doc":"- The telemetry event for message handles in the connection process should now\n  correctly publish the `:events` key, not the `:event` key\n- The telemetry guide now correctly states that the `:start_state` key holds\n  the connection state for handle-style events","ref":"changelog.html#fixed"},{"type":"extras","title":"Added - Changelog","doc":"- Published the connection state at the end of the message handle for\n  handle-type telemetry events in the connection process\n    - this is published in the `:end_state` key","ref":"changelog.html#added"},{"type":"extras","title":"0.3.0 - 2021-02-03 - Changelog","doc":"","ref":"changelog.html#0-3-0-2021-02-03"},{"type":"extras","title":"Added - Changelog","doc":"- `:telemetry` events are executed for each message handled by the\n  `Slipstream.Connection` process\n    - this should aide in debugging in scenarios in which one wants to see the\n      full event history of a connection","ref":"changelog.html#added"},{"type":"extras","title":"0.2.4 - 2021-02-02 - Changelog","doc":"","ref":"changelog.html#0-2-4-2021-02-02"},{"type":"extras","title":"Fixed - Changelog","doc":"- Emit `ChannelClosed` event on HTTP close from remote websocket server\n    - See [#11](https://github.com/NFIBrokerage/slipstream/issues/11)","ref":"changelog.html#fixed"},{"type":"extras","title":"0.2.3 - 2021-02-02 - Changelog","doc":"","ref":"changelog.html#0-2-3-2021-02-02"},{"type":"extras","title":"Changed - Changelog","doc":"- docs: the main page now shows the `Slipstream` module\n    - frankly, I just want people to see the Basic Usage section as soon as\n      possible so they get a taste for the library","ref":"changelog.html#changed"},{"type":"extras","title":"0.2.2 - 2021-02-01 - Changelog","doc":"","ref":"changelog.html#0-2-2-2021-02-01"},{"type":"extras","title":"Fixed - Changelog","doc":"- The connection now properly disconnects when the server's heartbeat replies\n  timeout with requests\n    - disconnects for this reason will send have `:heartbeat_timeout` as the\n      reason referenced in `c:Slipstream.handle_disconnect/2`","ref":"changelog.html#fixed"},{"type":"extras","title":"0.2.1 - 2021-02-01 - Changelog","doc":"","ref":"changelog.html#0-2-1-2021-02-01"},{"type":"extras","title":"Fixed - Changelog","doc":"- Connection now correctly emits ChannelClosed events for `:gun` events which\n  occur when a service's network access is terminated","ref":"changelog.html#fixed"},{"type":"extras","title":"0.2.0 - 2021-02-01 - Changelog","doc":"","ref":"changelog.html#0-2-0-2021-02-01"},{"type":"extras","title":"Added - Changelog","doc":"- Added a testing framework `Slipstream.SocketTest`","ref":"changelog.html#added"},{"type":"extras","title":"0.1.4 - 2021-01-31 - Changelog","doc":"","ref":"changelog.html#0-1-4-2021-01-31"},{"type":"extras","title":"Fixed - Changelog","doc":"- Published license to hex","ref":"changelog.html#fixed"},{"type":"extras","title":"0.1.3 - 2021-01-30 - Changelog","doc":"","ref":"changelog.html#0-1-3-2021-01-30"},{"type":"extras","title":"Changed - Changelog","doc":"- Widened dependency on `:gun` from `~> 1.3` to `~> 1.0`\n    - as far as I can tell, there aren't any real changes to the websocket\n      interface, so any of these versions should be OK","ref":"changelog.html#changed"},{"type":"extras","title":"0.1.2 - 2021-01-30 - Changelog","doc":"","ref":"changelog.html#0-1-2-2021-01-30"},{"type":"extras","title":"Added - Changelog","doc":"- More documentation\n    - including a Basic Usage section with a minimalistic example","ref":"changelog.html#added"},{"type":"extras","title":"0.1.1 - 2021-01-30 - Changelog","doc":"","ref":"changelog.html#0-1-1-2021-01-30"},{"type":"extras","title":"Removed - Changelog","doc":"- Removed direct dependency on `:telemetry`\n    - see #4","ref":"changelog.html#removed"},{"type":"extras","title":"0.1.0 - 2021-01-30 - Changelog","doc":"","ref":"changelog.html#0-1-0-2021-01-30"},{"type":"extras","title":"Added - Changelog","doc":"- Initial implementation and initial publish","ref":"changelog.html#added"},{"type":"extras","title":"0.0.0 - 2021-01-22 - Changelog","doc":"","ref":"changelog.html#0-0-0-2021-01-22"},{"type":"extras","title":"Added - Changelog","doc":"- This project was generated by Gaas\n\n<!-- # Generated by Elixir.Gaas.Generators.Simple.Library.Changelog -->","ref":"changelog.html#added"},{"type":"extras","title":"Telemetry","doc":"# Telemetry\n\nSlipstream emits telemetry events for both processes: the client process and\nthe connection process.\n\nTelemetry for the connection process is very low-level: information about\neach message and before-and-after states for the connection. This telemetry\naides in debugging failure-states of Slipstream and can be useful for\ndevelopment, but is likely not worth ingesting for average Slipstream users.\n\nClient process telemetry is similar to the telemetry emitted by\n`Phoenix.Channel`s and is recommended for consumption by average Slipstream\nusers.","ref":"telemetry.html"},{"type":"extras","title":"Client-process Telemetry - Telemetry","doc":"Clients emit telemetry in two categories:\n\n- basic channel connection and topic join events\n- callback events\n\nThe first emulate two-of-three events emitted by `Phoenix.Channel`s (at time\nof writing):\n\n- `[:slipstream, :client, :connect]` - dispatched at the end of a successful\n  client connection to a remote websocket host\n    - metadata\n      ```elixir\n      %{\n        start_time: DateTime.t(),\n        configuration: Slipstream.Configuration.t(),\n        socket: Slipstream.Socket.t(),\n        # emitted only on the stop event:\n        response_headers: [{String.t(), String.t()}]\n      }\n      ```\n- `[:slipstream, :client, :join]` - dispatched at the end of a successful\n  join onto a topic\n    - metadata\n      ```elixir\n      %{\n        start_time: DateTime.t(),\n        socket: Slipstream.Socket.t(),\n        topic: String.t(),\n        params: Slipstream.json_serializable(),\n        # emitted only on the stop event:\n        response: Slipstream.json_serializable()\n      }\n      ```\n\nThe second is emitted for any callback defined by the `Slipstream` module. Each\nof these events are emitted under the event name of\n`[:slipstream, :client, callback]`, where `callback` is any callback defined\nby `Slipstream`, e.g. `:handle_message`. The metadata for these events is as\nfollows:\n\n```elixir\n%{\n  client: module(),\n  callback: atom(),\n  arguments: [any()],\n  socket: Slipstream.Socket.t(),\n  start_time: DateTime.t(),\n  # emitted only on the stop event\n  response: any()\n}\n```\n\nNote that all client-process events emulate `:telemetry.span/3` in naming and\nin measurements, and in the case of callback events, the events are emitted\nby `:telemetry.span/3`. This means that each of the above-described events\nare actually prefixes, and each event name below represents messages of\n`event_prefix ++ [:start]` and `event_prefix ++ [:stop]`, with exceptions\nbeing emitted as `event_prefix ++ [:exception]`.\n\nClient authors are encouraged to only subscribe to the callbacks they are\ninterested in (e.g. `:handle_message` or `:handle_reply`).\n\nNote: events will _not_ be emitted for any messages handled by the `await_*`\nfamily of synchronous functions.","ref":"telemetry.html#client-process-telemetry"},{"type":"extras","title":"Connection-process Telemetry - Telemetry","doc":"Slipstream currently emits telemetry for each message received in its\n`Slipstream.Connection` process. This aides in low-level debugging when one\ndesires to see the entire event-history of a connection.\n\nA connection can be pieced together by the `:trace_id` and `:connection_id`\nkeys emitted in the metadata of these telemetry events. One\n`Slipstream.Connection` process is guaranteed to only use one `:trace_id` and\none `:connection_id`.\n\nThe trace-id acts as a identifier which groups together all requests belonging\nto that `Slipstream.Connection` process. Because the life-cycle of the GenServer\nwhich runs the `Slipstream.Connection` process and the actual connection to a\nremote websocket are the same, there doesn't appear to be much difference\nbetween the `:trace_id` and the `:connection_id`, however, the `:connection_id`\ncan be considered the parent-ID for all spans, where each subsequent message\nto the connection process is considered a span.\n\nThe `Slipstream.Connection` telemetry event prefixes are as follows:\n\n`[:slipstream, :connection, :connect]`: an event which is executed at the\nstart and stop of the connection process's life-cycle.\n\nThe metadata contains:\n\n```elixir\n%{\n  state: %Slipstream.Connection.State{},\n  connection_id: String.t(),\n  trace_id: String.t(),\n  # UTC date-time stamp of start-of-connection\n  start_time: DateTime.t()\n}\n```\n\n`[:slipstream, :connection, :handle]`: an event executed for each message\nreceived by the connection process.\n\nMetadata contains:\n\n```elixir\n%{\n  start_state: %Slipstream.Connection.State{},\n  connection_id: String.t(),\n  trace_id: String.t(),\n  # unique to this message handle\n  span_id: String.t(),\n  # UTC date-time stamp of start-of-handling\n  start_time: DateTime.t(),\n  raw_message: term(),\n  # included on the `event_prefix ++ [:stop]` events:\n  message: term(),\n  events: [%{type: atom(), attrs: map()}],\n  built_events: [struct()],\n  return: tuple(),\n  end_state: %Slipstream.Connection.State{}\n}\n```\n\nEach of these events follows the `:telemetry.span/3` pattern in measurements\nand in names: `event_prefix ++ [:start]` and `event_prefix ++ [:stop]` events\nare emitted using each of the above event prefixes.\n\nNote: connection telemetry will _not_ be emitted for any connections simulated\nin a `Slipstream.SocketTest`.","ref":"telemetry.html#connection-process-telemetry"},{"type":"extras","title":"Implementation","doc":"# Implementation\n\nSlipstream has some implementation decisions that may appear strange or\noverly-complicated on the surface. My opinion is that these help not only with\nusability but also test-ability. The test suite is rather compact: it covers\na great many cases for the small number of lines in which it is implemented.\n\nThe primary design decisions that enable this easy workflow are the hard\nboundary between the client and connection processes and the disambiguation\nof messages into events and commands:","ref":"implementation.html"},{"type":"extras","title":"Process Architecture - Implementation","doc":"Slipstream starts two processes to run a websocket connection:\n\n- the client process (modules with `use Slipstream`)\n- the connection process\n\nThe connection process is spawned and killed upon `Slipstream.connect/2`\nand `Slipstream.disconnect/1`, respectively. It directly interfaces with\nthe WebSocket connection via `Mint.WebSocket`.\n\nThe client process is simply a GenServer that has `c:GenServer.handle_info/2`\nclauses injected to handle slipstream events. The entire purpose of the client\nprocess is to provide a clean-feeling interface to the connection process. One\n_could_ (but is not recommended to-) forego the client process and deal with\nwebsockets synchronously with the `Slipstream.await_*` family of functions.","ref":"implementation.html#process-architecture"},{"type":"extras","title":"Events and Commands - Implementation","doc":"Slipstream separates out messages into two categories: events\nand commands. To some extent, Slipstream follows the [Event\nSourcing](https://martinfowler.com/eaaDev/EventSourcing.html) and [Command\nPattern](https://en.wikipedia.org/wiki/Command_pattern) design patterns with\nthis choice.\n\nOne might think to implement the communication between the [connection and\nclient processes](#process-architecture) by using data structures like\n`%Message{}` and `%Reply{}` and `%Heartbeat{}`, but it's not exactly clear\nwho should send which message to whom. Should the connection process be sending\n`%Message{}` to the server, or vice-versa? Could we re-use that data structure\nfor both directions?\n\nSlipstream's `Slipstream.Events` and `Slipstream.Commands` data structures\nkeep this boundary clear. The client process sends commands to the connection\nprocess when it seeks to change the connection state (join/leave) or push\na message. The connection process notifies the client of its status and how\nthe connection changes (whenever it changes meaningfully), with events. Events\ndeclare how the connection has changed, which informs the client process of how\nit should change the state of the `t:Slipstream.Socket.t/0` as the connection\nevolves.\n\nEvents are written in the past tense in the form of `NounVerbed`, as in\n`ThingHappened`, while commands are written in the form of `VerbNoun`, as in\n`DoThing`.","ref":"implementation.html#events-and-commands"},{"type":"extras","title":"Signatures - Implementation","doc":"Events arrive from the connection process in the format of\n\n```elixir\n{:__slipstream_event__, event_struct}\n```\n\nwhere `event_struct` is any struct from `Slipstream.Events`. And commands are\nsent to the connection process in the form of:\n\n```elixir\n{:__slipstream_command__, command_struct}\n```\n\nThe `Slipstream.Signatures` file defines two macros that wrap the above\npatterns. This allows one to both produce and match on a data structure\nand have confidence that the resulting data structure or match pattern will\nbe an event or command.\n\nThis pattern is not very revolutionary: it's really just an extension of DRY\ninto match patterns. It keeps the matches and routing functions clear of dirty\nimplementation details like 'exactly how the data structures are wrapped.'\n\nFor example, a `HeartbeatAcknowledged` event can be produced like so:\n\n```elixir\nalias Slipstream.Events\nimport Slipstream.Signatures, only: [event: 1]\n\nevent(%Events.HeartbeatAcknowledged{})\n#=> {:__slipstream_event__, %Slipstream.Events.HeartbeatAcknowledged{}}\n```\n\nAnd then later-on this data structure can be matched with any sort of\nclause like so:\n\n```elixir\nalias Slipstream.Events\nimport Slipstream.Signatures, only: [event: 1]\n\nmaybe_event = event(%Events.HeartbeatAcknowledged{})\n\ncase maybe_event do\n  event(%Events.HeartbeatAcknowledged{} = event) -> {:ok, event}\n  _ -> :error\nend\n#=> {:ok, %Events.HeartbeatAcknowledged{}}\n```\n\nSince the client and connection process communicate with each other though\nmessage sending and `GenServer.call/3`, these signature macros keep the\nmessages recognizable and enforce the communication boundary.\n\nI dub this pattern the \"signature pattern\" (but realize that I am probably\nnot the first to discover it).","ref":"implementation.html#signatures"},{"type":"extras","title":"The Token Pattern - Implementation","doc":"The `t:Slipstream.Socket.t/0` data structure is built to closely resemble the\n`t:Phoenix.Socket.t/0` structure, and feel similar in usage. These are both\nimplementations of a pattern coined by [`@rrrene`](https://github.com/rrrene):\nthe [token pattern](https://rrrene.org/2018/05/14/flow-elixir-designing-apis/).\n\nThe idea is fairly straight-forward: instead of passing around only the values\neach function needs, you pass around a structure which contains all information\navailable and provide a functional API to modify the internals (by producing a\nnew structure). These structures are called \"tokens.\"\n\nThe token pattern is particularly useful for writing flexible pipelines. E.g.\na piece of code that does something like this:\n\n```elixir\ndef do_work(request) do\n  request\n  |> decode_request()\n  |> hydrate_request_with_background_info()\n  |> start_computation_measurements()\n  |> do_the_computation()\n  |> end_computation_measurements()\n  |> send_the_result_somewhere()\nend\n```\n\nThese functions are all chained together, so it may seem perfectly appropriate\nto write the functions in a way that each consumes the direct output of the\npreceding function in the chain. But if we eventually need another `do_work/1`\nclause that handles a different sort of request and say, cuts out the\n`hydrate_request_with_background_info/1` step, we will likely have to refactor\nat least a few of the functions to properly emit and consume the data they need.\n\nInstead of taking and giving _only_ the information each function needs, each\nfunction should pass along a token structure containing all the information in\nthe request:\n\n```elixir\ndef do_work(request) do\n  %Token{initial_request: request}\n  |> decode_request() # returns a new %Token{}\n  |> hydrate_request_with_background_info() # returns a new %Token{}\n  ..\nend\n```\n\nIf we need to cut out or add new functions to the pipeline, we are less likely\nto need to refactor any of the component functions.\n\nThe usage of sockets in Slipstream is not so chain-y as the example, but the\nsocket token allows a clean API of functions that each take a socket and\nsome other information and emit a new socket token.\n\n```elixir\n@impl Slipstream\ndef handle_info({:join, topic}, socket) do\n  socket =\n    if connected?(socket) and not joined?(socket, topic) do\n      join(socket, topic)\n    else\n      socket\n    end\n\n  {:noreply, socket}\nend\n```","ref":"implementation.html#the-token-pattern"},{"type":"extras","title":"Examples","doc":"# Examples\n\nSlipstream provides a number of example socket client implementations as both a\ntutorial-like guide and a reference for common implementation patterns.\n\nEach of these examples is tested (via `mix test`), so each example is assured to\nbe in working order by every green check-mark from the CI.","ref":"examples.html"},{"type":"extras","title":"Format - Examples","doc":"Each example is contained in a directory in `./examples`.\nEach example contains a `README.md` with:\n\n- a description of the example and what it tries to solve\n- a tutorial-like listing of commits that created the example, along with\n  explanations of each step\n\nIn order to use the examples as a tutorial, try reading through each commit\nand reproducing the client in an example project.","ref":"examples.html#format"},{"type":"extras","title":"Directory - Examples","doc":"- Graceful Startup: a client that handles failures in configuration at start-up\n  time with graceful degradation\n- Repeater-style Client: a client which subscribes to a topic in another\n  service in order to re-publish each message as a broadcast in the client's\n  service's `Phoenix.Endpoint`\n- Rejoin on Reconnect: a client which re-joins all joined topics on a\n  reconnection after a disconnection\n- GenServer Capabilities: how to use the GenServer backend of Slipstream to\n  handle GenServer messaging","ref":"examples.html#directory"},{"type":"extras","title":"Graceful Startup","doc":"# Graceful Startup\n\nA client that gracefully handles errors in configuration at start-up time.\n\nAs the documentation for `c:GenServer.init/1` details, a GenServer may return\n`:ignore` instead of `{:ok, initial_state}` in its `c:GenServer.init/1`\ncallback. This will cause the client to\n\n> exit normally without entering the loop or calling `c:GenServer.terminate/2`.\n> If used when part of a supervision tree the parent supervisor will not\n> fail to start nor immediately try to restart the `GenServer`.\n\nThis makes it a good option for handling failures to start the client.\n`c:Slipstream.init/1` is mostly a wrapper around `c:GenServer.init/1`, so we\nmay return `:ignore` and see the same behavior.","ref":"graceful_startup_example.html"},{"type":"extras","title":"Tutorial - Graceful Startup","doc":"We start with an empty client module\n([`c5a08fd`](https://github.com/NFIBrokerage/slipstream/commit/c5a08fdcdb02bfe8721b4814f18e495942452426))\n\n```elixir\ndefmodule MyApp.GracefulStartupClient do\nend\n```\n\nAnd immediately fill out the Slipstream basics like a\n`start_link/1` implementation (so the module may be\nsupervised), and an invocation of `use Slipstream`\n([`8d44dd0`](https://github.com/NFIBrokerage/slipstream/commit/8d44dd057f6729b3d34787b85a57f0a968120946))\n\n```elixir\ndefmodule MyApp.GracefulStartupClient do\n  use Slipstream\n\n  @moduledoc \"...\"\n\n  def start_link(opts) do\n    Slipstream.start_link(__MODULE__, opts, name: __MODULE__)\n  end\nend\n```\n\nThen we add a basic implementation of the `c:Slipstream.init/1` callback\n([`a51e77c`](https://github.com/NFIBrokerage/slipstream/commit/a51e77c70b06086fe0ce5673d6650c0950cf30d8))\n\n```elixir\n@impl Slipstream\ndef init(_args) do\n  config = Application.fetch_env!(:my_app, __MODULE__)\n  {:ok, connect!(config)}\nend\n```\n\nBut note that this introduces two potential `raise`ing errors in our\n`c:Slipstream.init/1` callback:\n\n1. `Application.fetch_env!/2` will raise of the key-value pair is not defined\n   in configuration (`config/*.exs`)\n2. `Slipstream.connect!/2` will raise if the configuration passed as the first\n   argument is not valid according to `Slipstream.Configuration`.\n\nAnd also note that a `raise` in an `c:Slipstream.init/1` callback will fail\nthe start-up of the supervisor process. If this client is started in the\nApplication supervision tree (`lib/my_app/application.ex`), it will take\ndown the entire application on error.\n\nSo let's refactor this to make it a bit more safe! First, we switch\n`Application.fetch_env!/2` to its more graceful counterpart:\n`Application.fetch_env/2`, which returns `{:ok, config}` when\nthe configuration is defined and `:error` when it is not\n([`90ceb4a`](https://github.com/NFIBrokerage/slipstream/commit/90ceb4a5947b8e78c1ab9a1abfe4c6a55b1ab689))\n\n```elixir\n@impl Slipstream\ndef init(_args) do\n  with {:ok, config} <- Application.fetch_env(:slipstream, __MODULE__) do\n    {:ok, connect!(config)}\n  else\n    :error -> :ignore\n  end\nend\n```\n\nNow the client will attempt to connect only if the configuration\nis defined. But it can still fail, as we use the raising\n`Slipstream.connect!/2`.  We refactor that to `Slipstream.connect/2` in\n[`b3be445`](https://github.com/NFIBrokerage/slipstream/commit/b3be44525acfb6cda828a80d15e28d3351852540):\n\n```elixir\n@impl Slipstream\ndef init(_args) do\n  with {:ok, config} <- Application.fetch_env(:slipstream, __MODULE__),\n       {:ok, socket} <- connect(config) do\n    {:ok, socket}\n  else\n    :error -> :ignore\n    {:error, _reason} -> :ignore\n  end\nend\n```\n\nSo now in either of our failure-cases, the client will simply not start-up\ninstead of potentially crashing the entire application. This is an application\nof _graceful degradation_: in cases of system failure, we degrade our\nperformance instead of entirely giving up.\n\nIn this example, we go on to add helpful `Logger`\nmessages that declare when a failure case has been met\n([`09d1a83`](https://github.com/NFIBrokerage/slipstream/commit/09d1a83311b039d2dcd743a2a54cbba4287a21d2)).\nTo see the full example code, open up\n[`examples/graceful_startup/client.ex`](https://github.com/NFIBrokerage/slipstream/blob/main/examples/graceful_startup/client.ex).\nA small test suite can be found at\n[`test/slipstream/examples/graceful_startup_test.exs`](https://github.com/NFIBrokerage/slipstream/blob/main/test/slipstream/examples/graceful_startup_test.exs).","ref":"graceful_startup_example.html#tutorial"},{"type":"extras","title":"Repeater-style Client","doc":"# Repeater-style Client\n\nWe call any client that repeats messages from another service's endpoint a\n_repeater_ client.\n\nThis is a useful pattern when one wishes to ferry messages (without\nchanging them) from one `Phoenix.Endpoint` to another when the services\nare different, or when the services are the same but endpoint clustering is\nnot possible or desired.\n\nIn order to make this happen, we will create a client that\n\n- connects to the remove service\n- subscribes to the topic we wish to repeat\n- broadcasts messages as they happen to this service's endpoint","ref":"repeater_example.html"},{"type":"extras","title":"Tutorial - Repeater-style Client","doc":"We start with an empty client\n([`db79a32`](https://github.com/NFIBrokerage/slipstream/commit/db79a322e4b87ce4390fdb371076dcdbfb776ceb))\n\n```elixir\ndefmodule MyApp.RepeaterClient do\n  @moduledoc \"\"\"\n  A repeater-kind of client which re-broadcasts messages from another service\n  into this service's endpoint\n  \"\"\"\n\n  use Slipstream\n\n  def start_link(config) do\n    Slipstream.start_link(__MODULE__, config, name: __MODULE__)\n  end\n\n  @impl Slipstream\n  def init(config), do: {:ok, connect!(config)}\nend\n```\n\nThis client is pretty standard. It receives configuration from its supervisor,\nso it could be started in a `lib/my_app/application.ex` file with something\nlike:\n\n```elixir\ndef MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      {MyApp.RepeaterClient, uri: \"ws://service-b.local:4000/socket/websocket\"},\n      MyAppWeb.Endpoint\n    ]\n\n    Supervisor.start_link(children, strategy: :one_for_one)\n  end\nend\n```\n\nFirst thing's first, let's make sure we join our desired topic upon connection\nby adding a `c:Slipstream.handle_connect/1` implementation that calls\n`Slipstream.join/3`\n([`37b392e`](https://github.com/NFIBrokerage/slipstream/commit/37b392e823457f9542614da724cbf0c9e0181b73))\n\n```elixir\n@topic \"rooms:lobby\"\n\n@impl Slipstream\ndef handle_connect(socket), do: {:ok, join(socket, @topic)}\n```\n\nAnd let's add an empty implementation of `c:Slipstream.handle_message/4`\n([`3855594`](https://github.com/NFIBrokerage/slipstream/commit/38555946dfa583dc100084901e6a24e5467766da))\n\n```elixir\n@impl Slipstream\ndef handle_message(topic, event, payload, socket) do\n  {:ok, socket}\nend\n```\n\n`c:Slipstream.handle_message/4` is invoked whenever the remote server\nperforms a `Phoenix.Channel.push/3`, or whenever the topic to which the\nclient is subscribed receives a broadcast. Broadcasts are sent to all\nsubscribers of a topic without a `Phoenix.Channel` explicitly needing to use\n`Phoenix.Channel.push/3`, but the interface is the same for Slipstream clients:\n`c:Slipstream.handle_message/4` will be invoked with the broadcast.\n\nNow for the \"repeater\" part, we will publish these broadcasts or pushes onto\nour service's endpoint using `c:Phoenix.Endpoint.broadcast/3`\n([`01b5568`](https://github.com/NFIBrokerage/slipstream/commit/01b55688084b862fbe33df44dd66567766005835))\n\n```elixir\n@endpoint SlipstreamWeb.Endpoint\n\n@impl Slipstream\ndef handle_message(topic, event, payload, socket) do\n  @endpoint.broadcast(topic, event, payload)\n\n  {:ok, socket}\nend\n```\n\nAnd that's all! Each `Phoenix.Channel.push/3` or\n`c:Phoenix.Endpoint.broadcast/3` published on the topic will be (re)broadcasted\non the endpoint in the service of the repeater. See\n[`examples/repeater/client.ex`](https://github.com/NFIBrokerage/slipstream/blob/main/examples/repeater/client.ex)\nfor the full code of this example client and a small test case at\n[`test/slipstream/examples/repeater_test.exs`](https://github.com/NFIBrokerage/slipstream/blob/main/test/slipstream/examples/repeater_test.exs).","ref":"repeater_example.html#tutorial"},{"type":"extras","title":"Rejoin on Reconnect","doc":"# Rejoin on Reconnect\n\nSlipstream provides `rejoin/3` and `reconnect/1` to provide out-of-the-box\nretry mechanisms for re-establishing a client's functions after network\noutages.\n\nFor clients that\n\n- use the default implementation of `c:Slipstream.handle_disconnect/2`\n- call `Slipstream.connect/2` or `Slipstream.connect!/2` in\n  `c:Slipstream.init/1`\n- and call `Slipstream.join/3` in `c:Slipstream.handle_connect/1`\n\nThese retry mechanisms work without any extra effort. But for clients which\ndynamically join topics, you will have to take re-joins into account after a\ndisconnect, since that third condition about `Slipstream.join/3`ing may not\nbe true.\n\nConsider a client written like so:\n\n```elixir\ndefmodule MyClient do\n  use Slipstream\n\n  def join(topic) do\n    GenServer.cast(__MODULE__, {:join, topic})\n  end\n\n  def start_link(config) do\n    Slipstream.start_link(__MODULE__, config, name: __MODULE__)\n  end\n\n  @impl Slipstream\n  def init(config) do\n    {:ok, connect!(config)}\n  end\n\n  @impl Slipstream\n  def handle_cast({:join, topic}, socket) do\n    {:noreply, join(socket, topic)}\n  end\nend\n```\n\nThis client uses the default implementation of\n`c:Slipstream.handle_disconnect/2`, but does not join any topics in its\n`c:Slipstream.handle_connect/1` callback (since it uses the default\nimplementation). It doesn't know the topics it wishes to join ahead of time,\nso we say that this client joins topics _dynamically_.\n\nDynamic clients like this one will not automatically re-join after a\ndisconnect (however they will re-connect automatically due to the usage of the\ndefault `c:Slipstream.handle_disconnect/2` callback).","ref":"rejoin_on_reconnect_example.html"},{"type":"extras","title":"Tutorial - Rejoin on Reconnect","doc":"In this tutorial, we will start with the above client and refactor it so that\nit will gracefully re-join all topics when it reconnects after a disconnect.\n\nWe will start by modifying the `c:Slipstream.init/1` callback to store a list\nof topics in the socket assigns using `Slipstream.Socket.assign/3`\n([`2b0bd58`](https://github.com/NFIBrokerage/slipstream/commit/2b0bd582ff99d294274a323d8a902e101226c3be))\n\n```elixir\n@impl Slipstream\ndef init(config) do\n  socket =\n    config\n    |> connect!()\n    |> assign(:topics, [])\n    \n  {:ok, socket}\nend\n```\n\nAnd every time the client is told to join a topic with `MyClient.join/1`, we\nwant to update that `socket.assigns.topics` key to store the new topic. We\ndo that in the `c:Slipstream.handle_cast/2` callback with\n`Slipstream.Socket.update/3`\n([`70e802f`](https://github.com/NFIBrokerage/slipstream/pull/22/commits/70e802f7710a46d889939325ba7d62a680b60a96))\n\n```elixir\n@impl Slipstream\ndef handle_cast({:join, new_topic}, socket) do\n  socket =\n    socket\n    |> update(:topics, fn existing_topics -> [new_topic | existing_topics] end)\n    |> join(new_topic)\n\n  {:noreply, socket}\nend\n```\n\nSo now we have in `socket.assigns.topics` a list of all topics that we have\nattempted to join. (Curious how to store only the topics we have _successfully_\njoined? That is left as an exercise to the reader, with the solution written as\nan HTML comment at the bottom of this file.)\n\nNow we must implement a mechanism to re-join after a disconnect. More\naccurately, we must implement a mechanism to re-join after a re-connection.\nWhen a client is re-connected after a disconnect, the\n`c:Slipstream.handle_connect/1` callback is invoked again (it is first\ninvoked upon the first successful connection to the remote websocket server).\nThat's the perfect place to put our re-join mechanism\n([`70e802f`](https://github.com/NFIBrokerage/slipstream/pull/22/commits/70e802f7710a46d889939325ba7d62a680b60a96))\n\n```elixir\n@impl Slipstream\ndef handle_connect(socket) do\n  socket =\n    socket.assigns.topics\n    |> Enum.reduce(socket, fn topic, socket ->\n      case rejoin(socket, topic) do\n        {:ok, socket} -> socket\n        {:error, _reason} -> socket\n      end\n    end)\n\n  {:ok, socket}\nend\n```\n\nWhenever we connect, be it the first connect or a re-connection, we will\nattempt to `Slipstream.rejoin/3` all the topics in `socket.assigns.topic`.\nOn the initial connect this list is empty, so we keep the dynamic nature of\nonly joining topics on request (via `GenServer.cast/2`).\n\nThink that this should be the default behavior of a reconnection? Advocate for\nthis feature being built-in behavior in\n[#18](https://github.com/NFIBrokerage/slipstream/issues/18).\n\n<!--\n\nIn order to store only the topics we have _successfully_ joined, we simply\nmove the call to `Slipstream.Socket.update/3` into the\n`c:Slipstream.handle_join/3` callback implementation, defining it if it does\nnot yet exist. That callback is only invoked on a successful topic join, so\nadding the topic to the `socket.assigns.topics` list then will ensure that\nonly successfully-joined topics make their way into that key.\n\n-->","ref":"rejoin_on_reconnect_example.html#tutorial"},{"type":"extras","title":"GenServer Capabilities","doc":"# GenServer Capabilities\n\nSlipstream's callbacks are effectively a superset of `GenServer`, and in\nfact each module-based Slipstream socket client _is_ a `GenServer`. The\n`GenServer` abstraction provides a simple but powerful interface for message\nsending and internal state in a system with many actors.","ref":"gen_server_example.html"},{"type":"extras","title":"Tutorial - GenServer Capabilities","doc":"In this tutorial, we'll cover the basics of GenServer operations with\nSlipstream clients: `GenServer.cast/2`, `GenServer.call/3`, and sending\nmessages via `Kernel.send/2`.\n\nIf you haven't already acquainted yourself with GenServers and how to supervise\nthem, see the official Elixir guides:\n\n- [`GenServer`s](https://elixir-lang.org/getting-started/mix-otp/genserver.html)\n- [Supervision and `Application`](https://elixir-lang.org/getting-started/mix-otp/supervisor-and-application.html)\n\nLet's begin with a fresh client\n\n```elixir\ndefmodule MyApp.GenServerClient do\n  use Slipstream\n\n  def start_link(opts) do\n    Slipstream.start_link(__MODULE__, opts, name: __MODULE__)\n  end\n\n  @impl Slipstream\n  def init(opts) do\n    {:ok, connect!(opts)}\n  end\nend\n```\n\nAlready we are taking advantage of `Slipstream.start_link/3` (which is roughly\nthe same as `GenServer.start_link/3`) and `c:Slipstream.init/1` (which is\nroughly the same as `c:GenServer.init/1` except that you must return a\n`t:Slipstream.Socket.t/0`).\n\nLet's add a clause for handling `GenServer.cast/2`s\n([`8d5837b`](https://github.com/NFIBrokerage/slipstream/commit/8d5837b8bdc6801dd8538fe78aefa06631f07b21))\n\n```elixir\n@impl Slipstream\ndef handle_cast({:join, topic, params}, socket) do\n  {:noreply, join(socket, topic, params)}\nend\n```\n\nThis clause will accept casts telling the client to join a topic and perform\nthe join within the client's process. Notice how `c:Slipstream.handle_cast/2`\nlooks exactly the same as `c:GenServer.handle_cast/2`. The two are completely\ncompatible and all returns accepted by the GenServer callback are accepted by\nSlipstream as well.\n\nIf one were to ask the client to join a topic, they need only\n`GenServer.cast/2` it\n\n```elixir\niex> GenServer.cast(MyApp.GenServerClient, {:join, \"rooms:lobby\", %{}})\n:ok\n```\n\nNow let's try an implementation to handle `GenServer.call/3` requests. Calling\nis a synchronous action for which the caller will block until a reply is\nreceived. This is the only callback in the set of `Slipstream` callbacks which\ncan produce a reply-tuple (`{:reply, reply, socket}`). Let's start with\nsomething simple like a ping request\n([`fef68cc`](https://github.com/NFIBrokerage/slipstream/commit/fef68cce490b4dc02d21bc39e16d9ad12b47c5d4))\n\n```elixir\n@impl Slipstream\ndef handle_call(:ping, _from, socket) do\n  {:reply, :pong, socket}\nend\n```\n\nThis example doesn't really do very much. If we call our client with a\n`:ping` request, we expect a `:pong` result.\n\n```elixir\niex> GenServer.call(MyApp.GenServerClient, :ping)\n:pong\n```\n\nLet's try a more interesting `GenServer.call/3` feature: the client replying\nasynchronously. Surely the caller blocks while calling `GenServer.call/3`, but\nthe server which hears the call does not necessarily _have_ to respond in a\nblocking fashion. With a combination of a noreply tuple (`{:noreply, socket}`)\nand `GenServer.reply/2`, we can provide an API for synchronously requesting a\njoin\n([`d878f90`](https://github.com/NFIBrokerage/slipstream/commit/d878f90c9b8c0e5f3f0de001dc3881e8a46a78c4))\n\n```elixir\n@impl Slipstream\ndef handle_call({:join, topic, params}, from, socket) do\n  socket =\n    socket\n    |> assign(:join_request, from)\n    |> join(topic, params)\n\n  {:noreply, socket}\nend\n\n@impl Slipstream\ndef handle_join(_topic, response, socket) do\n  GenServer.reply(socket.assigns.join_request, {:ok, response})\n\n  {:ok, socket}\nend\n```\n\nAnd now we may call from another process\n\n```elixir\niex> GenServer.call(MyApp.GenServerClient, {:join, \"rooms:lobby\", %{}})\n{:ok, %{}}\n```\n\nAnd only receive a reply when the topic has been joined.\n\nFinally, the `c:GenServer.handle_info/2` callback provides a catch-all callback\nfor handling any messages sent to the server which are not calls or casts.\nThis callback works asynchronously like `c:GenServer.handle_cast/2`, but can\nbe triggered by another process performing a `Kernel.send/2` instead of a\n`GenServer.cast/2`.\n\nWe might write a client which uses `:timer.send_interval/2`,\n`Process.send_after/4`, or just plain `Kernel.send/2` to remind itself to do\nwork later, or another process may wish to send the client a message to\nalter its behavior. We'll take the approach of that last suggestion with this\ncallback implementation for `c:Slipstream.handle_info/2`\n([`63d3759`](https://github.com/NFIBrokerage/slipstream/commit/63d37591879ef2085c154f281357cad203cb6ba8))\n\n```elixir\n@impl Slipstream\ndef handle_info(:conclude_subscription, socket) do\n  {:stop, :normal, disconnect(socket)}\nend\n```\n\nAnd we'll revisit that `use Slipstream` invocation to add\n\n```elixir\nuse Slipstream, restart: :temporary\n```\n\nThis will ensure that when we tell the client to clean up and disconnect from\nits subscription that it should shutdown and not be restarted.\n\nWhen the client hears a message `:conclude_subscription`, it will disconnect\ngracefully and shut down the `GenServer` process. We can trigger this with\n\n```elixir\niex> MyApp.GenServerClient |> Process.whereis() |> send(:conclude_subscription)\n:conclude_subscription\n```","ref":"gen_server_example.html#tutorial"},{"type":"extras","title":"Bonus - GenServer Capabilities","doc":"As an added bonus, we'll discuss the feature of continues and how to use\n`c:Slipstream.handle_continue/2` to schedule a block of work to come next.\n\nSince OTP 21, the `c:GenServer.handle_continue/2` callback has provided a way\nto immediately schedule the next block of work without performing a\n`send(self(), {:next_work, params})` message.\n\nYou may wish to make use of this feature by scheduling the connection request\nto occur after `c:Slipstream.init/1`, soas the application supervision tree\nmay start up quickly, in the case where a client would have to do a large\namount of work before connecting.\n\n```elixir\n@impl Slipstream\ndef init(config) do\n  socket = new_socket() |> assign(:config, config)\n  {:ok, socket, {:continue, :connect}}\nend\n\n@impl Slipstream\ndef handle_continue(:connect, socket) do\n  # do some expensive work...\n  {:noreply, connect!(socket, socket.assigns.config)}\nend\n```\n\nThis will make the call to `Slipstream.connect!/2` occur _after_ the conclusion\nof the `c:Slipstream.init/1` callback.","ref":"gen_server_example.html#bonus"},{"type":"extras","title":"Scripting","doc":"# Scripting\n\nSometimes, it's really nice to safely play and experiment in your local\ndevelopment environment.\n\nIn this tutorial we'll cover how to use the synchronous `await_*` family of\nfunctions to write scripts for interacting with `Phoenix.Channel`s in an\nIEx session or in a `mix run script.exs` script. In this tutorial, unlike\nothers, we simply walk through a script from top to bottom, as opposed to\nmodifying an existing client.\n\nFor the sake of ease of development and for you, the reader, this repository\nruns a phoenix server with a socket endpoint on `localhost:4000` in dev-mode.\n\nTo prepare for this tutorial, clone the\n[`NFIBrokerage/slipstream`](https://github.com/NFIBrokerage) repository,\ninstall the necessary Elixir and Erlang versions with `asdf install` (see more\nabout [asdf](https://github.com/asdf-vm/asdf-elixir)). Then start an IEx server\nwith `MIX_ENV=dev iex -S mix`. The `.iex.exs` script in the root of this\nrepository will start the phoenix endpoint on `localhost:4000` as the IEx\nsession begins.\n\n```console\n$ iex -S mix\nErlang/OTP 23 [erts-11.1] [source] [64-bit] [smp:4:4] [ds:4:4:10] [async-threads:1] [hipe]\n\nInteractive Elixir (1.11.2) - press Ctrl+C to exit (type h() ENTER for help)\n20:13:38.963 [info] Running SlipstreamWeb.Endpoint with cowboy 2.8.0 at 0.0.0.0:4000 (http)\n20:13:38.967 [info] Access SlipstreamWeb.Endpoint at http://localhost:4000\niex(1)>\n```","ref":"scripting_example.html"},{"type":"extras","title":"Tutorial - Scripting","doc":"To begin, let's import some functions from Slipstream that we'll use to\nestablish and interact with our connection.\n\n```elixir\niex> import Slipstream\niex> import Slipstream.Socket\n```\n\nThe `Slipstream` module defines most of the functions we'll use to open and\ninteract with a connection while `Slipstream.Socket` provides helper functions\nto interact with our `socket` (a `t:Slipstream.Socket.t/0`).\n\nLet's connect!\n\n```elixir\niex> socket = connect!(uri: \"ws://localhost:4000/socket/websocket\") |> await_connect!\n#Slipstream.Socket \n```\n\nWe connect to the _websocket_ endpoint for `SlipstreamWeb.Endpoint`.\nThe endpoint has a `socket/3` definition like so:\n\n```elixir\nsocket(\"/socket\", SlipstreamWeb.UserSocket,\n  websocket: true,\n  longpoll: false\n)\n```\n\nMeaning that we can find the websocket endpoint at `/socket/websocket`. Note\nalso that we've changed the _scheme_ of the URI from `http` to `ws`.\n\nUpon connection we should also see a logger message from the\n`SlipstreamWeb.Endpoint` saying that a client has connected to the socket.\n\n```text\n20:20:59.874 [info] CONNECTED TO SlipstreamWeb.UserSocket in 60µs\n  Transport: :websocket\n  Serializer: Phoenix.Socket.V2.JSONSerializer\n  Parameters: %{\"vsn\" => \"2.0.0\"}\n```\n\nBut why do we need `await_connect!/1` in this scenario? The `connect/2`,\n`join/3`, `leave/2`, etc. functions are asynchronous requests to connect,\njoin, and leave, respectively. The `await_*` functions synchronously block\nuntil the responses to those requests are received.\n\nNow that we're connected, let's join a topic! `SlipstreamWeb.UserSocket`\ndefines a channel on the `\"rooms:lobby\"` topic like so\n\n```elixir\nchannel(\"rooms:lobby\", SlipstreamWeb.InteractiveChannel)\n```\n\nSo lets connect to that:\n\n```elixir\niex> topic = \"rooms:lobby\"\niex> socket = join(socket, topic, %{\"fizz\" => \"buzz\"}) |> await_join!(topic)\n#Slipstream.Socket \n```\n\nAnd we should see a message from the endpoint about the join:\n\n```\n20:38:35.208 [info] JOINED rooms:lobby in 16µs\n  Parameters: %{\"fizz\" => \"buzz\"}\n```\n\nAnd now let's push a message to a `c:Phoenix.Channel.handle_in/3` written like\nso\n\n```elixir\n# just swallows the request\ndef handle_in(\"quicksand\", _params, socket) do\n  {:noreply, socket}\nend\n```\n\n```elixir\niex> push!(socket, topic, \"quicksand\", %{\"a\" => \"b\"})\n\"2\"\n```\n\nWhy no `await_push` function for `push!/4`? And what's with the `\"2\"` as the\nreturn? `push!/4` (and `push/5`) are a bit different from the other functions.\nPushes are completely asynchronous as the `Phoenix.Channel` does not have to\nreply to any push. The `\"2\"` that got returned can be used to find replies,\nhowever.\n\nLet's push to a `c:Phoenix.Channel.handle_in/3` clause that _does_ reply\n\n```elixir\n# responds to the request\ndef handle_in(\"ping\", _params, socket) do\n  {:reply, {:ok, %{\"pong\" => \"pong\"}}, socket}\nend\n```\n\n```elixir\niex> push!(socket, topic, \"ping\", %{}) |> await_reply!()\n{:ok, %{\"ping\" => \"pong\"}}\n```\n\nNow we're using `await_reply!/1`, a function that synchronously waits for a\npush to receive a reply. But how does it know which push it is waiting for?\nThe `\"2\"` returned in the above `push!/4` example is what is called in the\nimplementation of Slipstream a \"ref.\" Refs are passed between the client and\nserver to signify which message a reply is referencing. E.g. if a join request\nsends a ref of `\"1\"`, the `Phoenix.Channel` will write a reply to that ref\ntelling the client whether or not the join has been accepted. `await_reply!/1`\nuses the ref returned from `push!/4` to find the reply to that push.\n\nNow we've pushed a message and seen a reply to a push, but what about receiving\na message that isn't a reply?\n\n```elixir\n# responds, but not with a reply\n# just an async send\ndef handle_in(\"push to me\", _params, socket) do\n  push(socket, \"foo\", %{\"bar\" => \"baz\"})\n\n  {:noreply, socket}\nend\n```\n\nThis `c:Phoenix.Channel.handle_in/3` clause will trigger a\n`Phoenix.Channel.push/3` to our client. In a module-based client, we would\ntypically handle this with `c:Slipstream.handle_message/4`, but how can we\nawait these messages in a script?\n\nLet push a message to this `handle_in`\n\n```elixir\niex> push!(socket, topic, \"push to me\", %{})\niex> flush\n{:__slipstream_event__,\n %Slipstream.Events.MessageReceived{\n   event: \"foo\",\n   payload: %{\"bar\" => \"baz\"},\n   topic: \"rooms:lobby\"\n }}\n```\n\nWe can see that the IEx process receives a message when we `flush/0`. We can\nawait and decode this message into a more digestible format with the\n`Slipstream.await_message!/4` macro. This macro takes patterns as its arguments\nfor the topic, event, and payload, so we can selectively match messages with\nmatch patterns.\n\n```elixir\niex> push!(socket, topic, \"push to me\", %{})\niex> await_message!(^topic, \"foo\", %{\"bar\" => \"baz\"})\n{\"rooms:lobby\", \"foo\", %{\"bar\" => \"baz\"}}\n```\n\nBut do we need to know the exact message we're awaiting ahead of time? No, we\ncan match all incoming messages like so\n\n```elixir\niex> push!(socket, topic, \"push to me\", %{})\niex> await_message!(_, _, _)\n{\"rooms:lobby\", \"foo\", %{\"bar\" => \"baz\"}}\n```\n\nFinally, let's leave the topic and disconnect, cleaning up our connection\ncleanly.\n\n```elixir\niex> socket = leave(socket, topic) |> await_leave!(topic)\n#Slipstream.Socket \niex> socket |> disconnect() |> await_disconnect!\n#Slipstream.Socket \n```","ref":"scripting_example.html#tutorial"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}